from landscape.tests.helpers import LandscapeTest

from landscape.lib.lsb_release import parse_lsb_release


class LsbReleaseTest(LandscapeTest):

    def test_parse_lsb_release(self):
        """
        L{parse_lsb_release} returns a C{dict} holding information from
        the given LSB release file.
        """
        lsb_release_filename = self.makeFile("DISTRIB_ID=Ubuntu\n"
                                             "DISTRIB_RELEASE=6.06\n"
                                             "DISTRIB_CODENAME=dapper\n"
                                             "DISTRIB_DESCRIPTION="
                                             "\"Ubuntu 6.06.1 LTS\"\n")

        self.assertEqual(parse_lsb_release(lsb_release_filename),
                         {"distributor-id": "Ubuntu",
                          "description": "Ubuntu 6.06.1 LTS",
                          "release": "6.06",
                          "code-name": "dapper"})

    def test_parse_lsb_release_with_missing_or_extra_fields(self):
        """
        L{parse_lsb_release} ignores lines not matching the map of
        known keys, and returns only keys with an actual value.
        """
        lsb_release_filename = self.makeFile("DISTRIB_ID=Ubuntu\n"
                                             "FOO=Bar\n")
        self.assertEqual(parse_lsb_release(lsb_release_filename),
                         {"distributor-id": "Ubuntu"})



import sys

from landscape.ui.tests.helpers import (
    ConfigurationProxyHelper, FakeGSettings, dbus_test_should_skip,
    dbus_skip_message, simulate_gtk_key_release, simulate_gtk_paste)

if not dbus_test_should_skip:
    from gi.repository import Gtk
    from landscape.ui.view.configuration import (
        ClientSettingsDialog, sanitise_host_name, is_valid_host_name)
    from landscape.ui.controller.configuration import ConfigController
    import landscape.ui.model.configuration.state
    from landscape.ui.model.configuration.state import (
        COMPUTER_TITLE, ConfigurationModel)
    from landscape.ui.model.configuration.uisettings import UISettings

from landscape.tests.helpers import LandscapeTest


class ViewFunctionsTest(LandscapeTest):

    def test_sanitise_host_name(self):
        """
        Test UI level host_name sanitation.
        """
        self.assertEqual("foo.bar", sanitise_host_name(" foo.bar"))
        self.assertEqual("foo.bar", sanitise_host_name("foo.bar "))
        self.assertEqual("foo.bar", sanitise_host_name(" foo.bar "))

    def test_is_valid_host_name_ok(self):
        """
        Test that valid host names cause L{is_valid_host_name} to return
        L{True}.
        """
        self.assertTrue(is_valid_host_name("a"))
        self.assertTrue(is_valid_host_name("a.b"))
        self.assertTrue(is_valid_host_name("a.b.c"))
        self.assertTrue(is_valid_host_name("stop-squark"))
        self.assertTrue(is_valid_host_name("stop-squark.teale.DE"))
        self.assertTrue(is_valid_host_name("a2.b3.c4"))

    def test_is_valid_host_name_bad(self):
        """
        Test that invalid host names cause L{is_valid_host_name} to return
        L{False}.
        """
        self.assertFalse(is_valid_host_name(".a"))
        self.assertFalse(is_valid_host_name("a."))
        self.assertFalse(is_valid_host_name("a b"))
        self.assertFalse(is_valid_host_name("a .b"))
        self.assertFalse(is_valid_host_name("a. b"))

    def test_is_valid_host_name_unicode(self):
        """
        Test that host names containing Unicode cause L{is_valid_host_name} to
        return L{False}.
        """
        self.assertFalse(is_valid_host_name(u"\xc3a"))

    if dbus_test_should_skip:
        skip = dbus_skip_message


class ConfigurationViewTest(LandscapeTest):

    helpers = [ConfigurationProxyHelper]

    def setUp(self):
        self.default_data = {"management-type": "canonical",
                             "computer-title": "",
                             "hosted-landscape-host": "",
                             "hosted-account-name": "",
                             "hosted-password": "",
                             "local-landscape-host": "",
                             "local-account-name": "",
                             "local-password": ""}

        self.config_string = (
            "[client]\n"
            "data_path = %s\n"
            "http_proxy = http://proxy.localdomain:3192\n"
            "tags = a_tag\n"
            "url = https://landscape.canonical.com/message-system\n"
            "account_name = foo\n"
            "registration_key = bar\n"
            "computer_title = baz\n"
            "https_proxy = https://proxy.localdomain:6192\n"
            "ping_url = http://landscape.canonical.com/ping\n" % sys.path[0])

        super(ConfigurationViewTest, self).setUp()
        landscape.ui.model.configuration.state.DEFAULT_DATA[COMPUTER_TITLE] \
            = "me.here.com"
        settings = FakeGSettings(data=self.default_data)
        self.uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy,
                                   uisettings=self.uisettings)
        self.controller = ConfigController(model)

    def run_gtk_eventloop(self):
        """Run the Gtk event loop until all events have been processed."""
        while Gtk.events_pending():
            Gtk.main_iteration()

    def assert_paste_data_saved(self, dialog, combo, widget, attribute):
        """
        Paste text into specified widget then verify data is saved.
        """
        # Switch to local mode
        dialog.use_type_combobox.set_active(combo)
        self.run_gtk_eventloop()

        simulate_gtk_paste(widget, "pasted text")
        self.run_gtk_eventloop()
        self.assertTrue(self.controller.is_modified)
        self.assertEqual("pasted text", getattr(self.controller, attribute))

        dialog.revert(None)
        self.run_gtk_eventloop()
        self.assertFalse(self.controller.is_modified)

    def test_init(self):
        """
        Test that we correctly initialise the L{ConfigurationView} correctly
        from the controller.
        """
        dialog = ClientSettingsDialog(self.controller)
        content_area = dialog.get_content_area()
        self.assertEqual("preferences-management-service",
                         dialog.get_default_icon_name())
        children = content_area.get_children()
        self.assertEqual(len(children), 2)
        box = children[0]
        self.assertIsInstance(box, Gtk.Box)
        self.assertEqual(1, dialog.use_type_combobox.get_active())

    def test_on_combobox_changed(self):
        """
        Test that changes to the active selection in L{use_type_combobox}
        result in the correct panel becoming active and visible.
        """
        dialog = ClientSettingsDialog(self.controller)
        iter = dialog.liststore.get_iter(0)
        no_service_frame = dialog.liststore.get(iter, 2)[0]
        iter = dialog.liststore.get_iter(1)
        hosted_service_frame = dialog.liststore.get(iter, 2)[0]
        iter = dialog.liststore.get_iter(2)
        local_service_frame = dialog.liststore.get(iter, 2)[0]

        self.assertEqual(1, dialog.use_type_combobox.get_active())
        [alignment] = dialog.register_button.get_children()
        [hbox] = alignment.get_children()
        [image, label] = hbox.get_children()

        self.run_gtk_eventloop()
        self.assertIs(hosted_service_frame, dialog.active_widget)
        self.assertEqual(dialog.REGISTER_BUTTON_TEXT, label.get_text())

        dialog.use_type_combobox.set_active(0)
        self.run_gtk_eventloop()
        self.assertIs(no_service_frame, dialog.active_widget)
        self.assertEqual(dialog.DISABLE_BUTTON_TEXT, label.get_text())

        dialog.use_type_combobox.set_active(2)
        self.run_gtk_eventloop()
        self.assertIs(local_service_frame, dialog.active_widget)
        self.assertEqual(dialog.REGISTER_BUTTON_TEXT, label.get_text())

    def test_modify(self):
        """
        Test that modifications to data in the UI are propagated to the
        controller.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        self.assertFalse(self.controller.is_modified)
        self.assertEqual(1, dialog.use_type_combobox.get_active())
        dialog.use_type_combobox.set_active(2)
        self.run_gtk_eventloop()
        self.assertTrue(self.controller.is_modified)
        dialog.revert(None)
        self.run_gtk_eventloop()
        self.assertFalse(self.controller.is_modified)
        simulate_gtk_key_release(dialog.hosted_account_name_entry, "A")
        self.run_gtk_eventloop()
        self.assertTrue(self.controller.is_modified)
        dialog.revert(None)
        self.run_gtk_eventloop()
        self.assertFalse(self.controller.is_modified)
        simulate_gtk_key_release(dialog.hosted_password_entry, "B")
        self.run_gtk_eventloop()
        self.assertTrue(self.controller.is_modified)
        dialog.revert(None)
        self.run_gtk_eventloop()
        self.assertFalse(self.controller.is_modified)
        simulate_gtk_key_release(dialog.local_landscape_host_entry, "C")
        self.run_gtk_eventloop()
        self.assertTrue(self.controller.is_modified)

    def test_modify_with_paste(self):
        """
        Non-keypress modifications to data in the UI are propagated to the
        controller.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        self.assertFalse(self.controller.is_modified)
        self.assertEqual(1, dialog.use_type_combobox.get_active())
        # Test hosted account name
        self.assert_paste_data_saved(dialog, 1,
                                     dialog.hosted_account_name_entry,
                                     "hosted_account_name")
        # Test hosted password
        self.assert_paste_data_saved(dialog, 1,
                                     dialog.hosted_password_entry,
                                     "hosted_password")
        # Test local hostname
        self.assert_paste_data_saved(dialog, 2,
                                     dialog.local_landscape_host_entry,
                                     "local_landscape_host")
        # Test local password
        self.assert_paste_data_saved(dialog, 2,
                                     dialog.local_password_entry,
                                     "local_password")

    def test_load_data_from_config(self):
        """
        Test that we load data into the appropriate entries from the
        configuration file.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.assertEqual(1, dialog.use_type_combobox.get_active())
        self.assertEqual("foo", dialog.hosted_account_name_entry.get_text())
        self.assertEqual("bar", dialog.hosted_password_entry.get_text())
        self.assertEqual("", dialog.local_landscape_host_entry.get_text())
        self.assertEqual("", dialog.local_password_entry.get_text())

    def test_revert(self):
        """
        Test that we can revert the UI values using the controller.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        self.assertEqual(1, dialog.use_type_combobox.get_active())
        self.assertEqual("foo", dialog.hosted_account_name_entry.get_text())
        self.assertEqual("bar", dialog.hosted_password_entry.get_text())
        dialog.use_type_combobox.set_active(2)
        dialog.local_landscape_host_entry.set_text("more.barn")
        self.run_gtk_eventloop()
        self.assertEqual("bar", dialog.hosted_password_entry.get_text())
        self.assertEqual("more.barn",
                         dialog.local_landscape_host_entry.get_text())
        dialog.revert(None)
        self.run_gtk_eventloop()
        self.assertEqual(1, dialog.use_type_combobox.get_active())
        self.assertEqual("foo", dialog.hosted_account_name_entry.get_text())
        self.assertEqual("bar", dialog.hosted_password_entry.get_text())

    def test_check_local_landscape_host_name_entry_ok(self):
        """
        Test that L{check_local_landscape_host_name_entry} returns L{True} when
        the input is a valid host name.
        """
        dialog = ClientSettingsDialog(self.controller)
        dialog.use_type_combobox.set_active(2)
        dialog.local_landscape_host_entry.set_text("foo.bar")
        self.assertTrue(dialog.check_local_landscape_host_name_entry())

    def test_check_local_landscape_host_name_entry_ok_not_recorded(self):
        """
        Test that L{check_local_landscape_host_name_entry} does not add the
        entry to L{ClientSettingsDialog._errored_entries} when the input is a
        valid host name.
        """
        dialog = ClientSettingsDialog(self.controller)
        dialog.use_type_combobox.set_active(2)
        dialog.local_landscape_host_entry.set_text("foo.bar")
        dialog.check_local_landscape_host_name_entry()
        self.assertEqual(0, len(dialog._errored_entries))

    def test_check_local_landscape_host_name_entry_bad_host_name(self):
        """
        Test that L{check_local_landscape_host_name_entry} returns L{False}
        when the input is not a valid host name.
        """
        dialog = ClientSettingsDialog(self.controller)
        dialog.use_type_combobox.set_active(2)
        dialog.local_landscape_host_entry.set_text("foo bar")
        self.assertFalse(dialog.check_local_landscape_host_name_entry())

    def test_check_local_landscape_host_name_entry_bad_recorded(self):
        """
        Test that L{check_local_landscape_host_name_entry} does add the
        entry to L{ClientSettingsDialog._errored_entries} when the input is not
        a valid host name.
        """
        dialog = ClientSettingsDialog(self.controller)
        dialog.use_type_combobox.set_active(2)
        dialog.local_landscape_host_entry.set_text("foo bar")
        dialog.check_local_landscape_host_name_entry()
        self.assertEqual(1, len(dialog._errored_entries))

    def test_check_local_landscape_host_name_entry_bad_error_type(self):
        """
        Test that L{check_local_landscape_host_name_entry} adds the correct
        error type to L{ClientSettingsDialog._validation_errors} when the input
        is not a valid host name.
        """
        dialog = ClientSettingsDialog(self.controller)
        dialog.use_type_combobox.set_active(2)
        dialog.local_landscape_host_entry.set_text("foo bar")
        dialog.check_local_landscape_host_name_entry()
        self.assertEqual(set([dialog.INVALID_HOST_NAME]),
                         dialog._validation_errors)

    def test_check_local_landscape_host_name_entry_unicode_in_host_name(self):
        """
        Test that L{check_local_landscape_host_name_entry} returns L{False}
        when the input contains Unicode.
        """
        dialog = ClientSettingsDialog(self.controller)
        dialog.use_type_combobox.set_active(2)
        dialog.local_landscape_host_entry.set_text(u"f\xc3.bar")
        self.assertFalse(dialog.check_local_landscape_host_name_entry())

    def test_check_local_landscape_host_name_entry_unicode_recorded(self):
        """
        Test that L{check_local_landscape_host_name_entry} does add the
        entry to L{ClientSettingsDialog._errored_entries} when the input
        contains Unicode.
        """
        dialog = ClientSettingsDialog(self.controller)
        dialog.use_type_combobox.set_active(2)
        dialog.local_landscape_host_entry.set_text(u"f\xc3.bar")
        dialog.check_local_landscape_host_name_entry()
        self.assertEqual(1, len(dialog._errored_entries))

    def test_check_local_landscape_host_name_entry_unicode_error_type(self):
        """
        Test that L{check_local_landscape_host_name_entry} adds the correct
        error type to L{ClientSettingsDialog._validation_errors} when the input
        contains Unicode.
        """
        dialog = ClientSettingsDialog(self.controller)
        dialog.use_type_combobox.set_active(2)
        dialog.local_landscape_host_entry.set_text(u"f\xc3.bar")
        dialog.check_local_landscape_host_name_entry()
        self.assertEqual(
            set([dialog.INVALID_HOST_NAME, dialog.UNICODE_IN_ENTRY]),
            dialog._validation_errors)

    def test_check_entry_ok(self):
        """
        Test that we return L{True} when the text of a L{Gtk.Entry} is valid
        input.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        dialog.use_type_combobox.set_active(1)
        self.run_gtk_eventloop()
        dialog.hosted_account_name_entry.set_text("Toodleoo")
        self.assertTrue(dialog.check_entry(dialog.hosted_account_name_entry))

    def test_check_entry_doesnt_record_entry_when_ok(self):
        """
        Test that, when the text of a L{Gtk.Entry} is valid nothing is added to
        L{ClientSettingsDialog._errored_entries}.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        dialog.use_type_combobox.set_active(1)
        self.run_gtk_eventloop()
        dialog.hosted_account_name_entry.set_text("Toodleoo")
        dialog.check_entry(dialog.hosted_account_name_entry)
        self.assertEqual(0, len(dialog._errored_entries))

    def test_check_entry_non_ascii(self):
        """
        Test that we return L{False} when the text of a L{Gtk.Entry} contains
        Unicode input.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        dialog.use_type_combobox.set_active(1)
        self.run_gtk_eventloop()
        dialog.hosted_account_name_entry.set_text(u"T\xc3dle\xc4")
        self.assertFalse(dialog.check_entry(dialog.hosted_account_name_entry))

    def test_check_entry_records_entry_when_non_ascii(self):
        """
        Test that, when the text of a L{Gtk.Entry} contains Unicode it is
        added to L{ClientSettingsDialog._errored_entries}.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        dialog.use_type_combobox.set_active(1)
        self.run_gtk_eventloop()
        dialog.hosted_account_name_entry.set_text(u"T\xc3dle\xc4")
        dialog.check_entry(dialog.hosted_account_name_entry)
        self.assertEqual(1, len(dialog._errored_entries))

    def test_dismiss_validation_errors_local(self):
        """
        Test that dismissing the validation errors tidies up indicators that
        have been set against local settings.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        dialog.use_type_combobox.set_active(1)
        self.run_gtk_eventloop()
        dialog.hosted_account_name_entry.set_text(u"T\xc3dle\xc4")
        dialog.hosted_password_entry.set_text(u"T\xc3dle\xc4")
        self.run_gtk_eventloop()
        dialog.validity_check()
        self.run_gtk_eventloop()
        self.assertEqual(2, len(dialog._errored_entries))
        [entry1, entry2] = dialog._errored_entries
        self.assertEqual(Gtk.STOCK_DIALOG_WARNING,
                         entry1.get_icon_stock(Gtk.EntryIconPosition.PRIMARY))
        self.assertEqual(Gtk.STOCK_DIALOG_WARNING,
                         entry2.get_icon_stock(Gtk.EntryIconPosition.PRIMARY))
        dialog.dismiss_infobar(None)
        self.run_gtk_eventloop()
        self.assertEqual(0, len(dialog._errored_entries))
        self.assertNotEqual(
            Gtk.STOCK_DIALOG_WARNING,
            entry1.get_icon_stock(Gtk.EntryIconPosition.PRIMARY))
        self.assertNotEqual(
            Gtk.STOCK_DIALOG_WARNING,
            entry2.get_icon_stock(Gtk.EntryIconPosition.PRIMARY))

    def test_dismiss_validation_errors_hosted(self):
        """
        Test that dismissing the validation errors tidies up indicators that
        have been set against hosted fields.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        dialog.use_type_combobox.set_active(2)
        self.run_gtk_eventloop()
        dialog.local_landscape_host_entry.set_text("dodgy as hell")
        self.run_gtk_eventloop()
        dialog.validity_check()
        self.run_gtk_eventloop()
        self.assertEqual(1, len(dialog._errored_entries))
        [entry1] = dialog._errored_entries
        self.assertEqual(Gtk.STOCK_DIALOG_WARNING,
                         entry1.get_icon_stock(Gtk.EntryIconPosition.PRIMARY))
        dialog.dismiss_infobar(None)
        self.run_gtk_eventloop()
        self.assertEqual(0, len(dialog._errored_entries))
        self.assertNotEqual(
            Gtk.STOCK_DIALOG_WARNING,
            entry1.get_icon_stock(Gtk.EntryIconPosition.PRIMARY))

    def test_validity_check_disabled(self):
        """
        Test that the L{validity_check} returns True when we disable landscape
        client.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        dialog.use_type_combobox.set_active(0)
        self.run_gtk_eventloop()
        self.assertTrue(dialog.validity_check())

    def test_validity_check_hosted(self):
        """
        Test that the L{validity_check} returns True when the hosted fields are
        valid.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        dialog.use_type_combobox.set_active(1)
        dialog.hosted_account_name_entry.set_text("Bob")
        dialog.hosted_password_entry.set_text("the builder")
        self.run_gtk_eventloop()
        self.assertTrue(dialog.validity_check())

    def test_validity_check_hosted_unicode(self):
        """
        Test that the L{validity_check} returns False when the hosted fields
        contain Unicode.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        dialog.use_type_combobox.set_active(1)
        dialog.hosted_account_name_entry.set_text(u"B\xc3b")
        self.run_gtk_eventloop()
        self.assertFalse(dialog.validity_check())

    def test_validity_check_local_ok(self):
        """
        Test that the L{validity_check} returns True when the local fields
        are valid.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        dialog.use_type_combobox.set_active(2)
        self.run_gtk_eventloop()
        dialog.local_landscape_host_entry.set_text("foo.bar")
        self.run_gtk_eventloop()
        self.assertTrue(dialog.validity_check())

    def test_validity_check_local_sanitisable(self):
        """
        Test that the L{validity_check} returns True when the local fields
        are valid after sanitation.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        dialog.use_type_combobox.set_active(2)
        dialog.local_landscape_host_entry.set_text(" foo.bar")
        self.run_gtk_eventloop()
        self.assertTrue(dialog.validity_check())
        dialog.local_landscape_host_entry.set_text("foo.bar ")
        self.run_gtk_eventloop()
        self.assertTrue(dialog.validity_check())

    def test_validity_check_local_invalid_host_name(self):
        """
        Test that the L{validity_check} returns False when the host name is
        invalid.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        dialog.use_type_combobox.set_active(2)
        dialog.local_landscape_host_entry.set_text("foo bar")
        self.run_gtk_eventloop()
        self.assertFalse(dialog.validity_check())

    def test_validity_check_local_unicode(self):
        """
        Test that the L{validity_check} returns False when the host name
        contains Unicode.
        """
        dialog = ClientSettingsDialog(self.controller)
        self.run_gtk_eventloop()
        dialog.use_type_combobox.set_active(2)
        dialog.local_landscape_host_entry.set_text(u"f\xc3.bar")
        self.run_gtk_eventloop()
        self.assertFalse(dialog.validity_check())

    if dbus_test_should_skip:
        skip = dbus_skip_message


class LocalConfigurationViewTest(LandscapeTest):

    helpers = [ConfigurationProxyHelper]

    def setUp(self):
        self.default_data = {"management-type": "LDS",
                             "computer-title": "",
                             "hosted-landscape-host": "",
                             "hosted-account-name": "",
                             "hosted-password": "",
                             "local-landscape-host": "",
                             "local-account-name": "",
                             "local-password": "manky"}

        self.config_string = (
            "[client]\n"
            "data_path = %s\n"
            "url = https://landscape.localdomain/message-system\n"
            "computer_title = baz\n"
            "ping_url = http://landscape.localdomain/ping\n" % sys.path[0])

        super(LocalConfigurationViewTest, self).setUp()
        landscape.ui.model.configuration.state.DEFAULT_DATA[COMPUTER_TITLE] \
            = "me.here.com"
        settings = FakeGSettings(data=self.default_data)
        self.uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy,
                                   uisettings=self.uisettings)
        self.controller = ConfigController(model)

    def test_init(self):
        """
        Test that we correctly initialise the L{ConfigurationView} correctly
        from the controller.
        """
        dialog = ClientSettingsDialog(self.controller)
        while Gtk.events_pending():
            Gtk.main_iteration()
        content_area = dialog.get_content_area()
        children = content_area.get_children()
        self.assertEqual(len(children), 2)
        box = children[0]
        self.assertIsInstance(box, Gtk.Box)
        self.assertEqual(2, dialog.use_type_combobox.get_active())

    def test_load_data_from_config(self):
        """
        Test that we load data into the appropriate entries from the
        configuration file.
        """
        dialog = ClientSettingsDialog(self.controller)
        while Gtk.events_pending():
            Gtk.main_iteration()
        self.assertEqual(2, dialog.use_type_combobox.get_active())
        self.assertEqual("", dialog.hosted_account_name_entry.get_text())
        self.assertEqual("", dialog.hosted_password_entry.get_text())
        self.assertEqual("landscape.localdomain",
                         dialog.local_landscape_host_entry.get_text())
        self.assertEqual("manky", dialog.local_password_entry.get_text())

    if dbus_test_should_skip:
        skip = dbus_skip_message



import unittest

from landscape.lib.scriptcontent import (build_script,
    generate_script_hash)


class ScriptContentTest(unittest.TestCase):

    def test_concatenate(self):
        self.assertEqual(build_script(u"/bin/sh", u"echo 1.0\n"),
                         "#!/bin/sh\necho 1.0\n")

    def test_concatenate_null_strings(self):
        self.assertEqual(build_script(None, None),
                         "#!\n")

    def test_generate_script_hash(self):
        hash1 = generate_script_hash("#!/bin/sh\necho 1.0\n")
        hash2 = generate_script_hash("#!/bin/sh\necho 1.0\n")
        hash3 = generate_script_hash("#!/bin/sh\necho 3.0\n")

        self.assertEqual(hash1, hash2)
        self.assertNotEqual(hash1, hash3)
        self.assertTrue(isinstance(hash1, str))



import os

from landscape.tests.helpers import LandscapeTest

from landscape.lib.vm_info import get_vm_info, get_container_info


class GetVMInfoTest(LandscapeTest):

    def setUp(self):
        super(GetVMInfoTest, self).setUp()
        self.root_path = self.makeDir()
        self.proc_path = self.makeDir(
            path=os.path.join(self.root_path, "proc"))
        self.sys_path = self.makeDir(path=os.path.join(self.root_path, "sys"))
        self.proc_sys_path = self.makeDir(
            path=os.path.join(self.proc_path, "sys"))

    def make_sys_vendor(self, content):
        """Create /sys/class/dmi/id/sys_vendor with the specified content."""
        dmi_path = os.path.join(self.root_path, "sys/class/dmi/id")
        self.makeDir(path=dmi_path)
        self.makeFile(dirname=dmi_path, basename="sys_vendor", content=content)

    def test_get_vm_info_empty_when_no_virtualization_is_found(self):
        """
        L{get_vm_info} should be empty when there's no virtualisation.
        """
        self.assertEqual(u"", get_vm_info(root_path=self.root_path))

    def test_get_vm_info_is_openvz_when_proc_vz_exists(self):
        """
        L{get_vm_info} should return 'openvz' when /proc/vz exists.
        """
        proc_vz_path = os.path.join(self.proc_path, "vz")
        self.makeFile(path=proc_vz_path, content="foo")

        self.assertEqual("openvz", get_vm_info(root_path=self.root_path))

    def test_get_vm_info_is_xen_when_proc_sys_xen_exists(self):
        """
        L{get_vm_info} should return 'xen' when /proc/sys/xen exists.
        """
        proc_sys_xen_path = os.path.join(self.proc_sys_path, "xen")
        self.makeFile(path=proc_sys_xen_path, content="foo")

        self.assertEqual("xen", get_vm_info(root_path=self.root_path))

    def test_get_vm_info_is_xen_when_sys_bus_xen_is_non_empty(self):
        """
        L{get_vm_info} should return 'xen' when /sys/bus/xen exists and has
        devices.
        """
        devices_xen_path = os.path.join(self.sys_path, "bus/xen/devices")
        self.makeDir(path=devices_xen_path)
        foo_devices_path = os.path.join(devices_xen_path, "foo")
        self.makeFile(path=foo_devices_path, content="bar")

        self.assertEqual("xen", get_vm_info(root_path=self.root_path))

    def test_get_vm_info_is_xen_when_proc_xen_exists(self):
        """
        L{get_vm_info} should return 'xen' when /proc/xen exists.
        """
        proc_xen_path = os.path.join(self.proc_path, "xen")
        self.makeFile(path=proc_xen_path, content="foo")

        self.assertEqual("xen", get_vm_info(root_path=self.root_path))

    def test_get_vm_info_is_empty_without_xen_devices(self):
        """
        L{get_vm_info} returns an empty string if the /sys/bus/xen/devices
        directory exists but doesn't contain any file.
        """
        devices_xen_path = os.path.join(self.sys_path, "bus/xen/devices")
        self.makeDir(path=devices_xen_path)

        self.assertEqual("", get_vm_info(root_path=self.root_path))

    def test_get_vm_info_with_bochs_sys_vendor(self):
        """
        L{get_vm_info} should return "kvm" when we detect the sys_vendor is
        Bochs.
        """
        self.make_sys_vendor("Bochs")
        self.assertEqual("kvm", get_vm_info(root_path=self.root_path))

    def test_get_vm_info_with_openstack_sys_vendor(self):
        """
        L{get_vm_info} should return "kvm" when we detect the sys_vendor is
        Openstack.
        """
        self.make_sys_vendor("OpenStack Foundation")
        self.assertEqual("kvm", get_vm_info(root_path=self.root_path))

    def test_get_vm_info_with_vmware_sys_vendor(self):
        """
        L{get_vm_info} should return "vmware" when we detect the sys_vendor is
        VMware Inc.
        """
        self.make_sys_vendor("VMware, Inc.")
        self.assertEqual("vmware", get_vm_info(root_path=self.root_path))

    def test_get_vm_info_with_virtualbox_sys_vendor(self):
        """
        L{get_vm_info} should return "virtualbox" when we detect the sys_vendor
        is innotek GmbH.
        """
        self.make_sys_vendor("innotek GmbH")
        self.assertEqual("virtualbox", get_vm_info(root_path=self.root_path))

    def test_get_vm_info_with_microsoft_sys_vendor(self):
        """
        L{get_vm_info} returns "hyperv" if the sys_vendor is Microsoft.
        """
        self.make_sys_vendor("Microsoft Corporation")
        self.assertEqual("hyperv", get_vm_info(root_path=self.root_path))

    def test_get_vm_info_with_kvm_on_other_architecture(self):
        """
        L{get_vm_info} returns 'kvm', if no sys_vendor is available but the
        model in /proc/cpuinfo contains 'emulated by qemu'.
        """
        cpuinfo_path = os.path.join(self.proc_path, "cpuinfo")
        cpuinfo = (
            "platform	: Some Machine\n"
            "model	: Some CPU (emulated by qemu)\n"
            "machine	: Some Machine (emulated by qemu)\n")
        self.makeFile(path=cpuinfo_path, content=cpuinfo)
        self.assertEqual("kvm", get_vm_info(root_path=self.root_path))

    def test_get_vm_info_with_other_vendor(self):
        """
        L{get_vm_info} should return an empty string when the sys_vendor is
        unknown.
        """
        self.make_sys_vendor("Some other vendor")
        self.assertEqual("", get_vm_info(root_path=self.root_path))


class GetContainerInfoTest(LandscapeTest):

    def test_no_container(self):
        """If not running in a container, an empty string is returned."""
        self.assertEqual("", get_container_info(path="/does/not/exist"))

    def test_in_container(self):
        """If running in a container, the container type is returned."""
        path = self.makeFile(content="lxc")
        self.assertEqual("lxc", get_container_info(path=path))

    def test_strip_newline(self):
        """The container type doesn't contain newlines."""
        path = self.makeFile(content="lxc\n")
        self.assertEqual("lxc", get_container_info(path=path))



import os
import json

from twisted.internet.defer import succeed

from landscape.lib.fs import touch_file
from landscape.tests.helpers import LandscapeTest, MonitorHelper
from landscape.monitor.cephusage import CephUsage


SAMPLE_OLD_TEMPLATE = (
    "   health HEALTH_WARN 6 pgs degraded; 6 pgs stuck "
    "unclean\n"
    "monmap e2: 3 mons at {server-269703f4-5217-495a-b7f2-b3b3473c1719="
    "10.55.60.238:6789/0,server-3f370698-f3b0-4cbe-8db9-a18e304c952b="
    "10.55.60.141:6789/0,server-f635fa07-e36f-453c-b3d5-b4ce86fbc6ff="
    "10.55.60.241:6789/0}, election epoch 8, quorum 0,1,2 "
    "server-269703f4-5217-495a-b7f2-b3b3473c1719,"
    "server-3f370698-f3b0-4cbe-8db9-a18e304c952b,"
    "server-f635fa07-e36f-453c-b3d5-b4ce86fbc6ff\n   "
    "osdmap e9: 3 osds: 3 up, 3 in\n    "
    "pgmap v114: 192 pgs: 186 active+clean, 6 active+degraded; "
    "0 bytes data, %s MB used, %s MB / %s MB avail\n   "
    "mdsmap e1: 0/0/1 up\n\n")

SAMPLE_NEW_TEMPLATE = (
    "health HEALTH_OK\n"
    "   monmap e2: 3 mons at {inst-007=192.168.64.139:6789/0,"
    "inst-008=192.168.64.140:6789/0,inst-009=192.168.64.141:6789/0}, "
    "election epoch 6, quorum 0,1,2 inst-007,inst-008,inst-009\n"
    "   osdmap e28: 3 osds: 3 up, 3 in\n"
    "    pgmap v193861: 208 pgs: 208 active+clean; 5514 MB data, %s MB used, "
    "%s MB / %s MB avail; 1739KB/s wr, 54op/s\n"
    "   mdsmap e1: 0/0/1 up\n")

SAMPLE_OUTPUT = SAMPLE_NEW_TEMPLATE % (4296, 53880, 61248)
SAMPLE_OLD_OUTPUT = SAMPLE_OLD_TEMPLATE % (4296, 53880, 61248)

SAMPLE_QUORUM = (
    '{ "election_epoch": 8,\n'
    '  "quorum": [\n'
    '        0,\n'
    '        1,\n'
    '        2],\n'
    '  "monmap": { "epoch": 2,\n'
    '      "fsid": "%s",\n'
    '      "modified": "2013-01-13 16:58:00.141737",\n'
    '      "created": "0.000000",\n'
    '      "mons": [\n'
    '            { "rank": 0,\n'
    '              "name": "server-1be72d64-0ff2-4ac1-ad13-1c06c8201011",\n'
    '              "addr": "10.55.60.188:6789\/0"},\n'
    '            { "rank": 1,\n'
    '              "name": "server-e847f147-ed13-46c2-8e6d-768aa32657ab",\n'
    '              "addr": "10.55.60.202:6789\/0"},\n'
    '            { "rank": 2,\n'
    '              "name": "server-3c831a0b-51d5-43a9-95d5-63644f0965cc",\n'
    '              "addr": "10.55.60.205:6789\/0"}]}}\n')

SAMPLE_QUORUM_OUTPUT = SAMPLE_QUORUM % "ecbb8960-0e21-11e2-b495-83a88f44db01"


class CephUsagePluginTest(LandscapeTest):
    helpers = [MonitorHelper]

    def setUp(self):
        super(CephUsagePluginTest, self).setUp()
        self.mstore = self.broker_service.message_store
        self.plugin = CephUsage(create_time=self.reactor.time)

    def test_wb_get_ceph_usage_if_command_not_found(self):
        """
        When the ceph command cannot be found or accessed, the
        C{_get_ceph_usage} method returns None.
        """
        self.plugin._get_status_command_output = lambda: succeed(None)
        self.monitor.add(self.plugin)

        self.assertIs(
            None, self.successResultOf(self.plugin._get_ceph_usage()))

    def test_wb_get_ceph_usage(self):
        """
        When the ceph command call returns output, the _get_ceph_usage method
        returns the percentage of used space.
        """
        self.plugin._get_status_command_output = lambda: succeed(SAMPLE_OUTPUT)
        self.monitor.add(self.plugin)

        self.assertEqual(
            0.12029780564263323,
            self.successResultOf(self.plugin._get_ceph_usage()))

    def test_wb_get_ceph_usage_old_format(self):
        """
        The _get_ceph_usage method understands command output in the "old"
        format (the output changed around version 0.56.1)
        """
        self.plugin._get_status_command_output = (
            lambda: succeed(SAMPLE_OLD_OUTPUT))
        self.monitor.add(self.plugin)

        self.assertEqual(
            0.12029780564263323,
            self.successResultOf(self.plugin._get_ceph_usage()))

    def test_wb_get_ceph_usage_empty_disk(self):
        """
        When the ceph command call returns output for empty disks, the
        _get_ceph_usage method returns 0.0 .
        """
        self.plugin._get_status_command_output = (
            lambda: succeed(SAMPLE_NEW_TEMPLATE % (0, 100, 100)))
        self.monitor.add(self.plugin)

        self.assertEqual(
            0.0, self.successResultOf(self.plugin._get_ceph_usage()))

    def test_wb_get_ceph_usage_full_disk(self):
        """
        When the ceph command call returns output for empty disks, the
        _get_ceph_usage method returns 1.0 .
        """
        self.plugin._get_status_command_output = (
            lambda: succeed(SAMPLE_NEW_TEMPLATE % (100, 0, 100)))

        self.monitor.add(self.plugin)
        self.assertEqual(
            1.0, self.successResultOf(self.plugin._get_ceph_usage()))

    def test_wb_get_ceph_usage_no_information(self):
        """
        When the ceph command outputs something that does not contain the
        disk usage information, the _get_ceph_usage method returns None.
        """
        output = "Blah\nblah"
        error = "Could not parse command output: '%s'" % output
        self.log_helper.ignore_errors(error)

        self.plugin._get_status_command_output = lambda: succeed(output)

        self.monitor.add(self.plugin)
        self.assertIs(
            None, self.successResultOf(self.plugin._get_ceph_usage()))

    def test_never_exchange_empty_messages(self):
        """
        The plugin will create a message with an empty
        C{ceph-usages} list when no previous data is available.  If an empty
        message is created during exchange, it should not be queued.
        """
        self.mstore.set_accepted_types(["ceph-usage"])
        self.monitor.add(self.plugin)

        self.monitor.exchange()
        self.assertEqual(0, len(self.mstore.get_pending_messages()))

    def test_exchange_messages(self):
        """
        The Ceph usage plugin queues message when manager.exchange()
        is called.
        """
        ring_id = "whatever"
        self.mstore.set_accepted_types(["ceph-usage"])

        self.plugin._ceph_usage_points = [(60, 1.0)]
        self.plugin._ceph_ring_id = ring_id
        self.monitor.add(self.plugin)

        self.monitor.exchange()
        self.assertMessages(self.mstore.get_pending_messages(),
                            [{"type": "ceph-usage",
                              "ceph-usages": [(60, 1.0)],
                              "ring-id": ring_id}])

    def test_create_message(self):
        """
        Calling create_message returns an expected message.
        """
        ring_id = "blah"
        self.plugin._ceph_usage_points = []
        self.plugin._ceph_ring_id = ring_id
        message = self.plugin.create_message()

        self.assertIn("type", message)
        self.assertEqual(message["type"], "ceph-usage")
        self.assertIn("ceph-usages", message)
        self.assertEqual(ring_id, message["ring-id"])
        ceph_usages = message["ceph-usages"]
        self.assertEqual(len(ceph_usages), 0)

        point = (60, 1.0)
        self.plugin._ceph_usage_points = [point]
        message = self.plugin.create_message()
        self.assertIn("type", message)
        self.assertEqual(message["type"], "ceph-usage")
        self.assertIn("ceph-usages", message)
        self.assertEqual(ring_id, message["ring-id"])
        ceph_usages = message["ceph-usages"]
        self.assertEqual(len(ceph_usages), 1)
        self.assertEqual([point], ceph_usages)

    def test_no_message_if_not_accepted(self):
        """
        Don't add any messages at all if the broker isn't currently accepting
        their type.
        """
        interval = 30
        monitor_interval = 300

        plugin = CephUsage(
            interval=interval, monitor_interval=monitor_interval,
            create_time=self.reactor.time)

        self.monitor.add(plugin)

        self.reactor.advance(monitor_interval * 2)
        self.monitor.exchange()

        self.mstore.set_accepted_types(["ceph-usage"])
        self.assertMessages(list(self.mstore.get_pending_messages()), [])

    def test_wb_get_ceph_ring_id(self):
        """
        When given a well formatted command output, the _get_ceph_ring_id()
        method returns the correct ring_id.
        """
        uuid = "i-am-a-uuid"
        self.plugin._get_quorum_command_output = (
            lambda: succeed(SAMPLE_QUORUM % uuid))
        self.assertEqual(
            uuid, self.successResultOf(self.plugin._get_ceph_ring_id()))

    def test_wb_get_ceph_ring_id_valid_json_no_information(self):
        """
        When the _get_quorum_command_output method returns something without
        the ring uuid information present but that is valid JSON, the
        _get_ceph_ring_id method returns None.
        """
        error = "Could not get ring_id from output: '{\"election_epoch\": 8}'."
        self.log_helper.ignore_errors(error)

        def return_output():
            # Valid JSON - just without the info we're looking for.
            data = {"election_epoch": 8}
            return succeed(json.dumps(data))

        self.plugin._get_quorum_command_output = return_output
        self.assertIs(
            None, self.successResultOf(self.plugin._get_ceph_ring_id()))

    def test_wb_get_ceph_ring_id_no_information(self):
        """
        When the _get_quorum_command_output method returns something without
        the ring uuid information present, the _get_ceph_ring_id method returns
        None.
        """
        error = "Could not get ring_id from output: 'Blah\nblah'."
        self.log_helper.ignore_errors(error)

        self.plugin._get_quorum_command_output = lambda: succeed("Blah\nblah")
        self.assertIs(
            None, self.successResultOf(self.plugin._get_ceph_ring_id()))

    def test_wb_get_ceph_ring_id_command_exception(self):
        """
        When the _get_quorum_command_output method returns None (if an
        exception happened for example), the _get_ceph_ring_id method
        returns None and logs no error.
        """
        self.plugin._get_quorum_command_output = lambda: succeed(None)
        self.assertIs(
            None, self.successResultOf(self.plugin._get_ceph_ring_id()))

    def test_plugin_run(self):
        """
        The plugin's run() method fills the _ceph_usage_points with
        accumulated samples after each C{interval} period.
        The _ceph_ring_id member of the plugin is also filled with the output
        of the _get_ceph_ring_id method.
        """
        monitor_interval = 300
        interval = monitor_interval
        plugin = CephUsage(
            interval=interval, monitor_interval=monitor_interval,
            create_time=self.reactor.time)

        uuid = "i-am-a-unique-snowflake"

        # The config file must be present for the plugin to run.
        ceph_client_dir = os.path.join(self.config.data_path, "ceph-client")
        ceph_conf = os.path.join(ceph_client_dir, "ceph.landscape-client.conf")
        os.mkdir(ceph_client_dir)
        touch_file(ceph_conf)
        plugin._ceph_config = ceph_conf

        plugin._get_quorum_command_output = (
            lambda: succeed(SAMPLE_QUORUM % uuid))
        plugin._get_status_command_output = (
            lambda: succeed(SAMPLE_NEW_TEMPLATE % (100, 0, 100)))
        self.monitor.add(plugin)

        self.reactor.advance(monitor_interval * 2)

        self.assertEqual([(300, 1.0), (600, 1.0)], plugin._ceph_usage_points)
        self.assertEqual(uuid, plugin._ceph_ring_id)

    def test_resynchronize_message_calls_reset_method(self):
        """
        If the reactor fires a "resynchronize" even the C{_reset}
        method on the ceph plugin object is called.
        """
        self.called = False

        def stub_reset():
            self.called = True

        self.plugin._reset = stub_reset

        self.monitor.add(self.plugin)

        self.reactor.fire("resynchronize")

        self.assertTrue(self.called)



import textwrap
from logging import getLogger
import math
import os

from twisted.python.failure import Failure

from landscape.lib.twisted_util import gather_results
from landscape.lib.log import log_failure
from landscape.plugin import PluginRegistry


class SysInfoPluginRegistry(PluginRegistry):
    """
    When the sysinfo plugin registry is run, it will run each of the
    registered plugins so that they get a chance to feed information
    into the registry.

    There are three kinds of details collected: headers, notes, and footnotes.

    They are presented to the user in a way similar to the following:

        Header1: Value1   Header3: Value3
        Header2: Value2   Header4: Value4

        => This is first note
        => This is the second note

        The first footnote.
        The second footnote.

    Headers are supposed to display information which is regularly
    available, such as the load and temperature of the system.  Notes
    contain eventual information, such as warnings of high temperatures,
    and low disk space.  Finally, footnotes contain pointers to further
    information such as URLs.
    """

    def __init__(self):
        super(SysInfoPluginRegistry, self).__init__()
        self._header_index = {}
        self._headers = []
        self._notes = []
        self._footnotes = []
        self._plugin_error = False

    def add_header(self, name, value):
        """Add a new information header to be displayed to the user.

        Each header name is only present once.  If a header is added
        multiple times, the last value added will be returned in
        the get_headers() call.

        Headers with value None are not returned by get_headers(), but
        they still allocate a position in the list.  This fact may be
        explored to create a deterministic ordering even when dealing
        with values obtained asynchornously.
        """
        index = self._header_index.get(name)
        if index is None:
            self._header_index[name] = len(self._headers)
            self._headers.append((name, value))
        else:
            self._headers[index] = (name, value)

    def get_headers(self):
        """Get all information headers to be displayed to the user.

        Headers which were added with value None are not included in
        the result.
        """
        return [pair for pair in self._headers if pair[1] is not None]

    def add_note(self, note):
        """Add a new eventual note to be shown up to the administrator."""
        self._notes.append(note)

    def get_notes(self):
        """Get all eventual notes to be shown up to the administrator."""
        return self._notes

    def add_footnote(self, note):
        """Add a new footnote to be shown up to the administrator."""
        self._footnotes.append(note)

    def get_footnotes(self):
        """Get all footnotes to be shown up to the administrator."""
        return self._footnotes

    def run(self):
        """Run all plugins, and return a deferred aggregating their results.

        This will call the run() method on each of the registered plugins,
        and return a deferred which aggregates each resulting deferred.
        """
        deferreds = []
        for plugin in self.get_plugins():
            try:
                result = plugin.run()
            except:
                self._log_plugin_error(Failure(), plugin)
            else:
                result.addErrback(self._log_plugin_error, plugin)
                deferreds.append(result)
        return gather_results(deferreds).addCallback(self._report_error_note)

    def _log_plugin_error(self, failure, plugin):
        self._plugin_error = True
        message = "%s plugin raised an exception." % plugin.__class__.__name__
        logger = getLogger("landscape-sysinfo")
        log_failure(failure, message, logger=logger)

    def _report_error_note(self, result):
        from landscape.sysinfo.deployment import get_landscape_log_directory
        if self._plugin_error:
            path = os.path.join(get_landscape_log_directory(), "sysinfo.log")
            self.add_note(
                "There were exceptions while processing one or more plugins. "
                "See %s for more information." % path)
        return result


def format_sysinfo(headers=(), notes=(), footnotes=(), width=80, indent="",
                   column_separator="   ", note_prefix="=> "):
    """Format sysinfo headers, notes and footnotes to be displayed.

    This function will format headers notes and footnotes in a way that
    looks similar to the following:

        Header1: Value1   Header3: Value3
        Header2: Value2   Header4: Value4

        => This is first note
        => This is the second note

        The first footnote.
        The second footnote.

    Header columns will be dynamically adjusted to conform to the size
    of header labels and values.
    """

    # Indentation spacing is easier to handle if we just take it off the width.
    width -= len(indent)

    headers_len = len(headers)
    value_separator = ": "

    # Compute the number of columns in the header.  To do that, we first
    # do a rough estimative of the maximum number of columns feasible,
    # and then we go back from there until we can fit things.
    min_length = width
    for header, value in headers:
        min_length = min(min_length, len(header)+len(value)+2) # 2 for ": "
    columns = int(math.ceil(float(width) /
                            (min_length + len(column_separator))))

    # Okay, we've got a base for the number of columns.  Now, since
    # columns may have different lengths, and the length of each column
    # will change as we compress headers in less and less columns, we
    # have to perform some backtracking to compute a good feasible number
    # of columns.
    while True:
        # Check if the current number of columns would fit in the screen.
        # Note that headers are indented like this:
        #
        #     Header:         First value
        #     Another header: Value
        #
        # So the column length is the sum of the widest header, plus the
        # widest value, plus the value separator.
        headers_per_column = int(math.ceil(headers_len / float(columns)))
        header_lengths = []
        total_length = 0
        for column in range(columns):
            # We must find the widest header and value, both to compute the
            # column length, and also to compute per-column padding when
            # outputing it.
            widest_header_len = 0
            widest_value_len = 0
            for row in range(headers_per_column):
                header_index = column * headers_per_column + row
                # There are potentially less headers in the last column,
                # so let's watch out for these here.
                if header_index < headers_len:
                    header, value = headers[header_index]
                    widest_header_len = max(widest_header_len, len(header))
                    widest_value_len = max(widest_value_len, len(value))

            if column > 0:
                # Account for the spacing between each column.
                total_length += len(column_separator)

            total_length += (widest_header_len + widest_value_len +
                             len(value_separator))

            # Keep track of these lengths for building the output later.
            header_lengths.append((widest_header_len, widest_value_len))

        if columns == 1 or total_length < width:
            # If there's just one column, or if we're within the requested
            # length, we're good to go.
            break

        # Otherwise, do the whole thing again with one less column.
        columns -= 1


    # Alright! Show time! Let's build the headers line by line.
    lines = []
    for row in range(headers_per_column):
        line = indent
        # Pick all columns for this line.  Note that this means that
        # for 4 headers with 2 columns, we pick header 0 and 2 for
        # the first line, since we show headers 0 and 1 in the first
        # column, and headers 2 and 3 in the second one.
        for column in range(columns):
            header_index = column * headers_per_column + row
            # There are potentially less headers in the last column, so
            # let's watch out for these here.
            if header_index < headers_len:
                header, value = headers[header_index]
                # Get the widest header/value on this column, for padding.
                widest_header_len, widest_value_len = header_lengths[column]
                if column > 0:
                    # Add inter-column spacing.
                    line += column_separator
                # And append the column to the current line.
                line += (header +
                         value_separator +
                         " " * (widest_header_len - len(header)) +
                         value)
                # If there are more columns in this line, pad it up so
                # that the next column's header is correctly aligned.
                if headers_len > (column+1) * headers_per_column + row:
                    line += " " * (widest_value_len - len(value))
        lines.append(line)

    if notes:
        if lines:
            # Some spacing between headers and notes.
            lines.append("")
        initial_indent = indent + note_prefix
        for note in notes:
            lines.extend(
                textwrap.wrap(note,
                              initial_indent=initial_indent,
                              subsequent_indent=" "*len(initial_indent),
                              width=width))

    if footnotes:
        if lines:
            lines.append("")
        lines.extend(indent + footnote for footnote in footnotes)

    return "\n".join(lines)



import os

from twisted.internet.defer import succeed

from landscape.monitor.aptpreferences import AptPreferences
from landscape.tests.helpers import LandscapeTest
from landscape.tests.helpers import MonitorHelper
from landscape.tests.mocker import ANY


class AptPreferencesTest(LandscapeTest):

    helpers = [MonitorHelper]

    def setUp(self):
        super(AptPreferencesTest, self).setUp()
        self.etc_apt_directory = self.makeDir()
        self.plugin = AptPreferences(self.etc_apt_directory)
        self.monitor.add(self.plugin)

    def test_get_data_without_apt_preferences_files(self):
        """
        L{AptPreferences.get_data} returns C{None} if no APT preferences file
        is detected.
        """
        self.assertIdentical(self.plugin.get_data(), None)

    def test_get_data_with_apt_preferences(self):
        """
        L{AptPreferences.get_data} includes the contents of the main APT
        preferences file.
        """
        preferences_filename = os.path.join(self.etc_apt_directory,
                                            "preferences")
        self.makeFile(path=preferences_filename, content="crap")
        self.assertEqual(self.plugin.get_data(),
                         {preferences_filename: "crap"})

    def test_get_data_with_empty_preferences_directory(self):
        """
        L{AptPreferences.get_data} returns C{None} if the APT preference
        directory is present but empty.
        """
        preferences_directory = os.path.join(self.etc_apt_directory,
                                             "preferences.d")
        self.makeDir(path=preferences_directory)
        self.assertIdentical(self.plugin.get_data(), None)

    def test_get_data_with_preferences_directory(self):
        """
        L{AptPreferences.get_data} includes the contents of all the file in the
        APT preferences directory.
        """
        preferences_directory = os.path.join(self.etc_apt_directory,
                                             "preferences.d")
        self.makeDir(path=preferences_directory)
        filename1 = self.makeFile(dirname=preferences_directory, content="foo")
        filename2 = self.makeFile(dirname=preferences_directory, content="bar")
        self.assertEqual(self.plugin.get_data(), {filename1: "foo",
                                                  filename2: "bar"})

    def test_get_data_with_one_big_file(self):
        """
        L{AptPreferences.get_data} truncates the contents of an APT preferences
        files bigger than the size limit.
        """
        preferences_filename = os.path.join(self.etc_apt_directory,
                                            "preferences")
        limit = self.plugin.size_limit
        self.makeFile(path=preferences_filename, content="a" * (limit + 1))
        self.assertEqual(self.plugin.get_data(), {
            preferences_filename: "a" * (limit - len(preferences_filename))})

    def test_get_data_with_many_big_files(self):
        """
        L{AptPreferences.get_data} truncates the contents of individual APT
        preferences files in the total size is bigger than the size limit.
        """
        preferences_directory = os.path.join(self.etc_apt_directory,
                                             "preferences.d")
        self.makeDir(path=preferences_directory)
        limit = self.plugin.size_limit
        filename1 = self.makeFile(dirname=preferences_directory,
                                  content="a" * (limit / 2))
        filename2 = self.makeFile(dirname=preferences_directory,
                                  content="b" * (limit / 2))
        self.assertEqual(self.plugin.get_data(),
                         {filename1: "a" * (limit / 2 - len(filename1)),
                          filename2: "b" * (limit / 2 - len(filename2))})

    def test_exchange_without_apt_preferences_data(self):
        """
        If the system has no APT preferences data, no message is sent.
        """
        self.mstore.set_accepted_types(["apt-preferences"])
        self.plugin.exchange()
        self.assertEqual(self.mstore.get_pending_messages(), [])

    def test_exchange(self):
        """
        If the system has some APT preferences data, a message of type
        C{apt-preferences} is sent. If the data then gets removed, a
        further message with the C{data} field set to C{None} is sent.
        """
        self.mstore.set_accepted_types(["apt-preferences"])
        main_preferences_filename = os.path.join(self.etc_apt_directory,
                                                 "preferences")
        self.makeFile(path=main_preferences_filename, content="crap")
        preferences_directory = os.path.join(self.etc_apt_directory,
                                             "preferences.d")
        self.makeDir(path=preferences_directory)
        sub_preferences_filename = self.makeFile(dirname=preferences_directory,
                                                 content="foo")
        self.plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(messages[0]["type"], "apt-preferences")
        self.assertEqual(messages[0]["data"],
                         {main_preferences_filename: u"crap",
                          sub_preferences_filename: u"foo"})
        for filename in messages[0]["data"]:
            self.assertTrue(isinstance(filename, unicode))

        # Remove all APT preferences data from the system
        os.remove(main_preferences_filename)
        os.remove(sub_preferences_filename)
        self.plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(messages[1]["type"], "apt-preferences")
        self.assertIdentical(messages[1]["data"], None)

    def test_exchange_only_once(self):
        """
        If the system has some APT preferences data, a message of type
        C{apt-preferences} is sent. If the data then gets removed, a
        further message with the C{data} field set to C{None} is sent.
        """
        self.mstore.set_accepted_types(["apt-preferences"])
        preferences_filename = os.path.join(self.etc_apt_directory,
                                            "preferences")
        self.makeFile(path=preferences_filename, content="crap")
        self.plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)

    def test_run(self):
        """
        If the server can accept them, the plugin should send
        C{apt-preferences} urgent messages.
        """
        self.mstore.set_accepted_types(["apt-preferences"])
        broker_mock = self.mocker.replace(self.remote)
        broker_mock.send_message(ANY, ANY, urgent=True)
        self.mocker.result(succeed(None))
        self.mocker.replay()
        preferences_filename = os.path.join(self.etc_apt_directory,
                                            "preferences")
        self.makeFile(path=preferences_filename, content="crap")
        self.plugin.run()
        self.mstore.set_accepted_types([])
        self.plugin.run()

    def test_resynchronize(self):
        """
        The "resynchronize" reactor message cause the plugin to send fresh
        data.
        """
        preferences_filename = os.path.join(self.etc_apt_directory,
                                            "preferences")
        self.makeFile(path=preferences_filename, content="crap")
        self.mstore.set_accepted_types(["apt-preferences"])
        self.plugin.run()
        self.reactor.fire("resynchronize", scopes=["package"])
        self.plugin.run()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 2)



import urlparse
import logging
import time
import sys
import os
import glob
import apt_pkg

from twisted.internet.defer import Deferred, succeed

from landscape.lib.sequenceranges import sequence_to_ranges
from landscape.lib.twisted_util import gather_results, spawn_process
from landscape.lib.fetch import fetch_async
from landscape.lib.fs import touch_file
from landscape.lib import bpickle

from landscape.package.taskhandler import (
    PackageTaskHandlerConfiguration, PackageTaskHandler, run_task_handler)
from landscape.package.store import UnknownHashIDRequest, FakePackageStore


HASH_ID_REQUEST_TIMEOUT = 7200
MAX_UNKNOWN_HASHES_PER_REQUEST = 500


class PackageReporterConfiguration(PackageTaskHandlerConfiguration):
    """Specialized configuration for the Landscape package-reporter."""

    def make_parser(self):
        """
        Specialize L{Configuration.make_parser}, adding options
        reporter-specific options.
        """
        parser = super(PackageReporterConfiguration, self).make_parser()
        parser.add_option("--force-apt-update", default=False,
                          action="store_true",
                          help="Force running apt-update.")
        return parser


class PackageReporter(PackageTaskHandler):
    """Report information about the system packages.

    @cvar queue_name: Name of the task queue to pick tasks from.
    """
    config_factory = PackageReporterConfiguration

    queue_name = "reporter"

    apt_update_filename = "/usr/lib/landscape/apt-update"
    sources_list_filename = "/etc/apt/sources.list"
    sources_list_directory = "/etc/apt/sources.list.d"
    _session_id = None
    _got_task = False

    def run(self):
        self._got_task = False

        result = Deferred()
        # Set us up to communicate properly
        result.addCallback(lambda x: self.get_session_id())

        result.addCallback(lambda x: self.run_apt_update())

        # If the appropriate hash=>id db is not there, fetch it
        result.addCallback(lambda x: self.fetch_hash_id_db())

        # Attach the hash=>id database if available
        result.addCallback(lambda x: self.use_hash_id_db())

        # Now, handle any queued tasks.
        result.addCallback(lambda x: self.handle_tasks())

        # Then, remove any expired hash=>id translation requests.
        result.addCallback(lambda x: self.remove_expired_hash_id_requests())

        # After that, check if we have any unknown hashes to request.
        result.addCallback(lambda x: self.request_unknown_hashes())

        # Finally, verify if we have anything new to report to the server.
        result.addCallback(lambda x: self.detect_changes())

        result.callback(None)
        return result

    def send_message(self, message):
        return self._broker.send_message(
            message, self._session_id, True)

    def fetch_hash_id_db(self):
        """
        Fetch the appropriate pre-canned database of hash=>id mappings
        from the server. If the database is already present, it won't
        be downloaded twice.

        The format of the database filename is <uuid>_<codename>_<arch>,
        and it will be downloaded from the HTTP directory set in
        config.package_hash_id_url, or config.url/hash-id-databases if
        the former is not set.

        Fetch failures are handled gracefully and logged as appropriate.
        """

        def fetch_it(hash_id_db_filename):

            if hash_id_db_filename is None:
                # Couldn't determine which hash=>id database to fetch,
                # just ignore the failure and go on
                return

            if os.path.exists(hash_id_db_filename):
                # We don't download twice
                return

            base_url = self._get_hash_id_db_base_url()
            if not base_url:
                logging.warning("Can't determine the hash=>id database url")
                return

            # Cast to str as pycurl doesn't like unicode
            url = str(base_url + os.path.basename(hash_id_db_filename))

            def fetch_ok(data):
                hash_id_db_fd = open(hash_id_db_filename, "w")
                hash_id_db_fd.write(data)
                hash_id_db_fd.close()
                logging.info("Downloaded hash=>id database from %s" % url)

            def fetch_error(failure):
                exception = failure.value
                logging.warning("Couldn't download hash=>id database: %s" %
                                str(exception))

            result = fetch_async(url,
                                 cainfo=self._config.get("ssl_public_key"))
            result.addCallback(fetch_ok)
            result.addErrback(fetch_error)

            return result

        result = self._determine_hash_id_db_filename()
        result.addCallback(fetch_it)
        return result

    def _get_hash_id_db_base_url(self):

        base_url = self._config.get("package_hash_id_url")

        if not base_url:

            if not self._config.get("url"):
                # We really have no idea where to download from
                return None

            # If config.url is http://host:123/path/to/message-system
            # then we'll use http://host:123/path/to/hash-id-databases
            base_url = urlparse.urljoin(self._config.url.rstrip("/"),
                                        "hash-id-databases")

        return base_url.rstrip("/") + "/"

    def _apt_sources_have_changed(self):
        """Return a boolean indicating if the APT sources were modified."""
        from landscape.monitor.packagemonitor import PackageMonitor

        filenames = []

        if os.path.exists(self.sources_list_filename):
            filenames.append(self.sources_list_filename)

        if os.path.exists(self.sources_list_directory):
            filenames.extend(
                [os.path.join(self.sources_list_directory, filename) for
                 filename in os.listdir(self.sources_list_directory)])

        for filename in filenames:
            seconds_since_last_change = (
                time.time() - os.path.getmtime(filename))
            if seconds_since_last_change < PackageMonitor.run_interval:
                return True

        return False

    def _apt_update_timeout_expired(self, interval):
        """Check if the apt-update timeout has passed."""
        if os.path.exists(self.update_notifier_stamp):
            stamp = self.update_notifier_stamp
        elif os.path.exists(self._config.update_stamp_filename):
            stamp = self._config.update_stamp_filename
        else:
            return True

        last_update = os.stat(stamp).st_mtime
        return (last_update + interval) < time.time()

    def run_apt_update(self):
        """Run apt-update and log a warning in case of non-zero exit code.

        @return: a deferred returning (out, err, code)
        """
        if (self._config.force_apt_update or self._apt_sources_have_changed()
            or self._apt_update_timeout_expired(
                self._config.apt_update_interval)):

            result = spawn_process(self.apt_update_filename)

            def callback((out, err, code)):
                accepted_apt_errors = (
                    "Problem renaming the file /var/cache/apt/srcpkgcache.bin",
                    "Problem renaming the file /var/cache/apt/pkgcache.bin")

                touch_file(self._config.update_stamp_filename)
                logging.debug(
                    "'%s' exited with status %d (out='%s', err='%s')" % (
                        self.apt_update_filename, code, out, err))

                if code != 0:
                    logging.warning("'%s' exited with status %d (%s)" % (
                        self.apt_update_filename, code, err))

                    # Errors caused by missing cache files are acceptable, as
                    # they are not an issue for the lists update process.
                    # These errors can happen if an 'apt-get clean' is run
                    # while 'apt-get update' is running.
                    for message in accepted_apt_errors:
                        if message in err:
                            out, err, code = "", "", 0
                            break

                elif not self._facade.get_channels():
                    code = 1
                    err = ("There are no APT sources configured in %s or %s." %
                           (self.sources_list_filename,
                            self.sources_list_directory))

                deferred = self._broker.call_if_accepted(
                    "package-reporter-result", self.send_result, code, err)
                deferred.addCallback(lambda ignore: (out, err, code))
                return deferred

            return result.addCallback(callback)

        else:
            logging.debug("'%s' didn't run, update interval has not passed" %
                          self.apt_update_filename)
            return succeed(("", "", 0))

    def send_result(self, code, err):
        """
        Report the package reporter result to the server in a message.
        """
        message = {
            "type": "package-reporter-result",
            "code": code,
            "err": err}
        return self.send_message(message)

    def handle_task(self, task):
        message = task.data
        if message["type"] == "package-ids":
            self._got_task = True
            return self._handle_package_ids(message)
        if message["type"] == "resynchronize":
            self._got_task = True
            return self._handle_resynchronize()

    def _handle_package_ids(self, message):
        unknown_hashes = []

        try:
            request = self._store.get_hash_id_request(message["request-id"])
        except UnknownHashIDRequest:
            # We've lost this request somehow.  It will be re-requested later.
            return succeed(None)

        hash_ids = {}

        for hash, id in zip(request.hashes, message["ids"]):
            if id is None:
                unknown_hashes.append(hash)
            else:
                hash_ids[hash] = id

        self._store.set_hash_ids(hash_ids)

        logging.info("Received %d package hash => id translations, %d hashes "
                     "are unknown." % (len(hash_ids), len(unknown_hashes)))

        if unknown_hashes:
            result = self._handle_unknown_packages(unknown_hashes)
        else:
            result = succeed(None)

        # Remove the request if everything goes well.
        result.addCallback(lambda x: request.remove())

        return result

    def _handle_resynchronize(self):
        self._store.clear_available()
        self._store.clear_available_upgrades()
        self._store.clear_installed()
        self._store.clear_locked()

        # Don't clear the hash_id_requests table because the messages
        # associated with the existing requests might still have to be
        # delivered, and if we clear the table and later create a new request,
        # that new request could get the same id of one of the deleted ones,
        # and when the pending message eventually gets delivered the reporter
        # would think that the message is associated to the newly created
        # request, as it has the same id has the deleted request the message
        # actually refers to. This would cause the ids in the message to be
        # possibly mapped to the wrong hashes.
        #
        # This problem would happen for example when switching the client from
        # one Landscape server to another, because the uuid-changed event would
        # cause a resynchronize task to be created by the monitor. See #417122.

        return succeed(None)

    def _handle_unknown_packages(self, hashes):

        self._facade.ensure_channels_reloaded()

        hashes = set(hashes)
        added_hashes = []
        packages = []
        for package in self._facade.get_packages():
            hash = self._facade.get_package_hash(package)
            if hash in hashes:
                added_hashes.append(hash)
                skeleton = self._facade.get_package_skeleton(package)
                packages.append({"type": skeleton.type,
                                 "name": skeleton.name,
                                 "version": skeleton.version,
                                 "section": skeleton.section,
                                 "summary": skeleton.summary,
                                 "description": skeleton.description,
                                 "size": skeleton.size,
                                 "installed-size": skeleton.installed_size,
                                 "relations": skeleton.relations})

        if packages:
            logging.info("Queuing messages with data for %d packages to "
                         "exchange urgently." % len(packages))

            message = {"type": "add-packages", "packages": packages}

            result = self._send_message_with_hash_id_request(message,
                                                             added_hashes)
        else:
            result = succeed(None)

        return result

    def remove_expired_hash_id_requests(self):
        now = time.time()
        timeout = now - HASH_ID_REQUEST_TIMEOUT

        def update_or_remove(is_pending, request):
            if is_pending:
                # Request is still in the queue.  Update the timestamp.
                request.timestamp = now
            elif request.timestamp < timeout:
                # Request was delivered, and is older than the threshold.
                request.remove()

        results = []
        for request in self._store.iter_hash_id_requests():
            if request.message_id is None:
                # May happen in some rare cases, when a send_message() is
                # interrupted abruptly.  If it just fails normally, the
                # request is removed and so we don't get here.
                request.remove()
            else:
                result = self._broker.is_message_pending(request.message_id)
                result.addCallback(update_or_remove, request)
                results.append(result)

        return gather_results(results)

    def request_unknown_hashes(self):
        """Detect available packages for which we have no hash=>id mappings.

        This method will verify if there are packages that APT knows
        about but for which we don't have an id yet (no hash => id
        translation), and deliver a message (unknown-package-hashes)
        to request them.

        Hashes previously requested won't be requested again, unless they
        have already expired and removed from the database.
        """
        self._facade.ensure_channels_reloaded()

        unknown_hashes = set()

        for package in self._facade.get_packages():
            hash = self._facade.get_package_hash(package)
            if self._store.get_hash_id(hash) is None:
                unknown_hashes.add(self._facade.get_package_hash(package))

        # Discard unknown hashes in existent requests.
        for request in self._store.iter_hash_id_requests():
            unknown_hashes -= set(request.hashes)

        if not unknown_hashes:
            result = succeed(None)
        else:
            unknown_hashes = sorted(unknown_hashes)
            unknown_hashes = unknown_hashes[:MAX_UNKNOWN_HASHES_PER_REQUEST]

            logging.info("Queuing request for package hash => id "
                         "translation on %d hash(es)." % len(unknown_hashes))

            message = {"type": "unknown-package-hashes",
                       "hashes": unknown_hashes}

            result = self._send_message_with_hash_id_request(message,
                                                             unknown_hashes)

        return result

    def _send_message_with_hash_id_request(self, message, unknown_hashes):
        """Create a hash_id_request and send message with "request-id"."""
        request = self._store.add_hash_id_request(unknown_hashes)
        message["request-id"] = request.id
        result = self.send_message(message)

        def set_message_id(message_id):
            request.message_id = message_id

        def send_message_failed(failure):
            request.remove()
            return failure

        return result.addCallbacks(set_message_id, send_message_failed)

    def detect_changes(self):
        """Detect all changes concerning packages.

        If some changes were detected with respect to our last run, then an
        event of type 'package-data-changed' will be fired in the broker
        reactor.
        """

        def changes_detected(result):
            if result:
                # Something has changed, notify the broker.
                return self._broker.fire_event("package-data-changed")

        deferred = self.detect_packages_changes()
        return deferred.addCallback(changes_detected)

    def detect_packages_changes(self):
        """
        Check if any information regarding packages have changed, and if so
        compute the changes and send a signal.
        """
        if self._got_task or self._package_state_has_changed():
            return self._compute_packages_changes()
        else:
            return succeed(None)

    def _package_state_has_changed(self):
        """
        Detect changes in the universe of known packages.

        This uses the state of packages in /var/lib/dpkg/state and other files
        and simply checks whether they have changed using their "last changed"
        timestamp on the filesystem.

        @return True if the status changed, False otherwise.
        """
        stamp_file = self._config.detect_package_changes_stamp
        if not os.path.exists(stamp_file):
            return True

        status_file = apt_pkg.config.find_file("dir::state::status")
        lists_dir = apt_pkg.config.find_dir("dir::state::lists")
        files = [status_file, lists_dir]
        files.extend(glob.glob("%s/*Packages" % lists_dir))

        last_checked = os.stat(stamp_file).st_mtime
        for f in files:
            last_changed = os.stat(f).st_mtime
            if last_changed >= last_checked:
                return True
        return False

    def _compute_packages_changes(self):
        """Analyse changes in the universe of known packages.

        This method will verify if there are packages that:

        - are now installed, and were not;
        - are now available, and were not;
        - are now locked, and were not;
        - were previously available but are not anymore;
        - were previously installed but are not anymore;
        - were previously locked but are not anymore;

        Additionally it will report package locks that:

        - are now set, and were not;
        - were previously set but are not anymore;

        In all cases, the server is notified of the new situation
        with a "packages" message.

        @return: A deferred resulting in C{True} if package changes were
            detected with respect to the previous run, or C{False} otherwise.
        """
        self._facade.ensure_channels_reloaded()

        old_installed = set(self._store.get_installed())
        old_available = set(self._store.get_available())
        old_upgrades = set(self._store.get_available_upgrades())
        old_locked = set(self._store.get_locked())

        current_installed = set()
        current_available = set()
        current_upgrades = set()
        current_locked = set()

        for package in self._facade.get_packages():
            hash = self._facade.get_package_hash(package)
            id = self._store.get_hash_id(hash)
            if id is not None:
                if self._facade.is_package_installed(package):
                    current_installed.add(id)
                    if self._facade.is_package_available(package):
                        current_available.add(id)
                else:
                    current_available.add(id)

                # Are there any packages that this package is an upgrade for?
                if self._facade.is_package_upgrade(package):
                    current_upgrades.add(id)

        for package in self._facade.get_locked_packages():
            hash = self._facade.get_package_hash(package)
            id = self._store.get_hash_id(hash)
            if id is not None:
                current_locked.add(id)

        new_installed = current_installed - old_installed
        new_available = current_available - old_available
        new_upgrades = current_upgrades - old_upgrades
        new_locked = current_locked - old_locked

        not_installed = old_installed - current_installed
        not_available = old_available - current_available
        not_upgrades = old_upgrades - current_upgrades
        not_locked = old_locked - current_locked

        message = {}
        if new_installed:
            message["installed"] = \
                list(sequence_to_ranges(sorted(new_installed)))
        if new_available:
            message["available"] = \
                list(sequence_to_ranges(sorted(new_available)))
        if new_upgrades:
            message["available-upgrades"] = \
                list(sequence_to_ranges(sorted(new_upgrades)))
        if new_locked:
            message["locked"] = \
                list(sequence_to_ranges(sorted(new_locked)))

        if not_installed:
            message["not-installed"] = \
                list(sequence_to_ranges(sorted(not_installed)))
        if not_available:
            message["not-available"] = \
                list(sequence_to_ranges(sorted(not_available)))
        if not_upgrades:
            message["not-available-upgrades"] = \
                list(sequence_to_ranges(sorted(not_upgrades)))
        if not_locked:
            message["not-locked"] = \
                list(sequence_to_ranges(sorted(not_locked)))

        if not message:
            return succeed(False)

        message["type"] = "packages"
        result = self.send_message(message)

        logging.info("Queuing message with changes in known packages: "
                     "%d installed, %d available, %d available upgrades, "
                     "%d locked, %d not installed, %d not available, "
                     "%d not available upgrades, %d not locked."
                     % (len(new_installed), len(new_available),
                        len(new_upgrades), len(new_locked),
                        len(not_installed), len(not_available),
                        len(not_upgrades), len(not_locked)))

        def update_currently_known(result):
            if new_installed:
                self._store.add_installed(new_installed)
            if not_installed:
                self._store.remove_installed(not_installed)
            if new_available:
                self._store.add_available(new_available)
            if new_locked:
                self._store.add_locked(new_locked)
            if not_available:
                self._store.remove_available(not_available)
            if new_upgrades:
                self._store.add_available_upgrades(new_upgrades)
            if not_upgrades:
                self._store.remove_available_upgrades(not_upgrades)
            if not_locked:
                self._store.remove_locked(not_locked)
            # Something has changed wrt the former run, let's update the
            # timestamp and return True.
            stamp_file = self._config.detect_package_changes_stamp
            touch_file(stamp_file)
            return True

        result.addCallback(update_currently_known)

        return result


class FakeGlobalReporter(PackageReporter):
    """
    A standard reporter, which additionally stores messages sent into its
    package store.
    """

    package_store_class = FakePackageStore

    def send_message(self, message):
        self._store.save_message(message)
        return super(FakeGlobalReporter, self).send_message(message)


class FakeReporter(PackageReporter):
    """
    A fake reporter which only sends messages previously stored by a
    L{FakeGlobalReporter}.
    """

    package_store_class = FakePackageStore
    global_store_filename = None

    def run(self):
        result = succeed(None)

        result.addCallback(lambda x: self.get_session_id())

        # If the appropriate hash=>id db is not there, fetch it
        result.addCallback(lambda x: self.fetch_hash_id_db())

        result.addCallback(lambda x: self._store.clear_tasks())

        # Finally, verify if we have anything new to send to the server.
        result.addCallback(lambda x: self.send_pending_messages())

        return result

    def send_pending_messages(self):
        """
        As the last callback of L{PackageReporter}, sends messages stored.
        """
        if self.global_store_filename is None:
            self.global_store_filename = os.environ["FAKE_PACKAGE_STORE"]
        if not os.path.exists(self.global_store_filename):
            return succeed(None)
        message_sent = set(self._store.get_message_ids())
        global_store = FakePackageStore(self.global_store_filename)
        all_message_ids = set(global_store.get_message_ids())
        not_sent = all_message_ids - message_sent
        deferred = succeed(None)
        got_type = set()
        if not_sent:
            messages = global_store.get_messages_by_ids(not_sent)
            sent = []
            for message_id, message in messages:
                message = bpickle.loads(str(message))
                if message["type"] not in got_type:
                    got_type.add(message["type"])
                    sent.append(message_id)
                    deferred.addCallback(
                        lambda x, message=message: self.send_message(message))
            self._store.save_message_ids(sent)
        return deferred


def main(args):
    if "FAKE_GLOBAL_PACKAGE_STORE" in os.environ:
        return run_task_handler(FakeGlobalReporter, args)
    elif "FAKE_PACKAGE_STORE" in os.environ:
        return run_task_handler(FakeReporter, args)
    else:
        return run_task_handler(PackageReporter, args)


def find_reporter_command():
    dirname = os.path.dirname(os.path.abspath(sys.argv[0]))
    return os.path.join(dirname, "landscape-package-reporter")



import time

try:
    import sqlite3
except ImportError:
    from pysqlite2 import dbapi2 as sqlite3

from landscape.tests.helpers import LandscapeTest

from landscape.broker.exchangestore import ExchangeStore


class ExchangeStoreTest(LandscapeTest):
    """Unit tests for the C{ExchangeStore}."""

    def setUp(self):
        super(ExchangeStoreTest, self).setUp()

        self.filename = self.makeFile()
        self.store1 = ExchangeStore(self.filename)
        self.store2 = ExchangeStore(self.filename)

    def test_add_message_context(self):
        """Adding a message context works correctly."""
        now = time.time()
        self.store1.add_message_context(123, 'abc', 'change-packages')

        db = sqlite3.connect(self.store2._filename)
        cursor = db.cursor()
        cursor.execute(
            "SELECT operation_id, secure_id, message_type, timestamp "
            "FROM message_context WHERE operation_id=?", (123,))
        results = cursor.fetchall()
        self.assertEqual(1, len(results))
        [row] = results
        self.assertEqual(123, row[0])
        self.assertEqual('abc', row[1])
        self.assertEqual('change-packages', row[2])
        self.assertTrue(row[3] > now)

    def test_add_message_context_with_duplicate_operation_id(self):
        """Only one message context with a given operation-id is permitted."""
        self.store1.add_message_context(123, "abc", "change-packages")
        self.assertRaises(
            (sqlite3.IntegrityError, sqlite3.OperationalError),
            self.store1.add_message_context, 123, "def", "change-packages")

    def test_get_message_context(self):
        """
        Accessing a C{MessageContext} with an existing C{operation-id} works.
        """
        now = time.time()
        self.store1.add_message_context(234, 'bcd', 'change-packages')
        context = self.store2.get_message_context(234)
        self.assertEqual(234, context.operation_id)
        self.assertEqual('bcd', context.secure_id)
        self.assertEqual('change-packages', context.message_type)
        self.assertTrue(context.timestamp > now)

    def test_get_message_context_with_nonexistent_operation_id(self):
        """Attempts to access a C{MessageContext} with a non-existent
        C{operation-id} result in C{None}."""
        self.assertIs(None, self.store1.get_message_context(999))

    def test_message_context_remove(self):
        """C{MessageContext}s are deleted correctly."""
        context = self.store1.add_message_context(
            345, 'opq', 'change-packages')
        context.remove()
        self.assertIs(None, self.store1.get_message_context(345))

    def test_all_operation_ids_for_empty_database(self):
        """
        Calling C{all_operation_ids} on an empty database returns an empty
        list.
        """
        self.assertEqual([], self.store1.all_operation_ids())

    def test_all_operation_ids(self):
        """C{all_operation_ids} works correctly."""
        self.store1.add_message_context(456, 'cde', 'change-packages')
        self.assertEqual([456], self.store2.all_operation_ids())
        self.store2.add_message_context(567, 'def', 'change-packages')
        self.assertEqual([456, 567], self.store1.all_operation_ids())



import pwd
import os
import sys
import tempfile
import stat

from twisted.internet.defer import gatherResults, succeed, fail
from twisted.internet.error import ProcessDone
from twisted.python.failure import Failure

from landscape import VERSION
from landscape.lib.fetch import HTTPCodeError
from landscape.lib.persist import Persist
from landscape.manager.scriptexecution import (
    ScriptExecutionPlugin, ProcessTimeLimitReachedError, PROCESS_FAILED_RESULT,
    UBUNTU_PATH, get_user_info, UnknownInterpreterError, UnknownUserError,
    FETCH_ATTACHMENTS_FAILED_RESULT)
from landscape.manager.manager import SUCCEEDED, FAILED
from landscape.tests.helpers import (
    LandscapeTest, ManagerHelper, StubProcessFactory, DummyProcess)
from landscape.tests.mocker import ANY, ARGS


def get_default_environment():
    username = pwd.getpwuid(os.getuid())[0]
    uid, gid, home = get_user_info(username)
    return  {
        "PATH": UBUNTU_PATH,
        "USER": username,
        "HOME": home}


class RunScriptTests(LandscapeTest):

    helpers = [ManagerHelper]

    def setUp(self):
        super(RunScriptTests, self).setUp()
        self.plugin = ScriptExecutionPlugin()
        self.manager.add(self.plugin)

    def test_basic_run(self):
        """
        The plugin returns a Deferred resulting in the output of basic
        commands.
        """
        result = self.plugin.run_script("/bin/sh", "echo hi")
        result.addCallback(self.assertEqual, "hi\n")
        return result

    def test_other_interpreter(self):
        """Non-shell interpreters can be specified."""
        result = self.plugin.run_script("/usr/bin/python", "print 'hi'")
        result.addCallback(self.assertEqual, "hi\n")
        return result

    def test_other_interpreter_env(self):
        """
        Non-shell interpreters don't have their paths set by the shell, so we
        need to check that other interpreters have environment variables set.
        """
        result = self.plugin.run_script(
            sys.executable,
            "import os\nprint os.environ")

        def check_environment(results):
            for string in get_default_environment():
                self.assertIn(string, results)

        result.addCallback(check_environment)
        return result

    def test_server_supplied_env(self):
        """
        Server-supplied environment variables are merged with default
        variables then passed to script.
        """
        server_supplied_env = {"DOG": "Woof", "CAT": "Meow"}
        result = self.plugin.run_script(
            sys.executable,
            "import os\nprint os.environ",
            server_supplied_env=server_supplied_env)

        def check_environment(results):
            for string in get_default_environment():
                self.assertIn(string, results)
            for name, value in server_supplied_env.items():
                self.assertIn(name, results)
                self.assertIn(value, results)

        result.addCallback(check_environment)
        return result

    def test_server_supplied_env_overrides_client(self):
        """
        Server-supplied environment variables override client default
        values if the server provides them.
        """
        server_supplied_env = {"PATH": "server-path", "USER": "server-user",
                               "HOME": "server-home"}
        result = self.plugin.run_script(
            sys.executable,
            "import os\nprint os.environ",
            server_supplied_env=server_supplied_env)

        def check_environment(results):
            for name, value in server_supplied_env.items():
                self.assertIn(name, results)
                self.assertIn(value, results)

        result.addCallback(check_environment)
        return result

    def test_concurrent(self):
        """
        Scripts run with the ScriptExecutionPlugin plugin are run concurrently.
        """
        fifo = self.makeFile()
        os.mkfifo(fifo)
        self.addCleanup(os.remove, fifo)
        # If the first process is blocking on a fifo, and the second process
        # wants to write to the fifo, the only way this will complete is if
        # run_script is truly async
        d1 = self.plugin.run_script("/bin/sh", "cat " + fifo)
        d2 = self.plugin.run_script("/bin/sh", "echo hi > " + fifo)
        d1.addCallback(self.assertEqual, "hi\n")
        d2.addCallback(self.assertEqual, "")
        return gatherResults([d1, d2])

    def test_accented_run_in_code(self):
        """
        Scripts can contain accented data both in the code and in the
        result.
        """
        accented_content = u"\N{LATIN SMALL LETTER E WITH ACUTE}"
        result = self.plugin.run_script(
            u"/bin/sh", u"echo %s" % (accented_content,))
        result.addCallback(
            self.assertEqual, "%s\n" % (accented_content.encode("utf-8"),))
        return result

    def test_accented_run_in_interpreter(self):
        """
        Scripts can also contain accents in the interpreter.
        """
        accented_content = u"\N{LATIN SMALL LETTER E WITH ACUTE}"
        result = self.plugin.run_script(
            u"/bin/echo %s" % (accented_content,), u"")

        def check(result):
            self.assertTrue(
                "%s " % (accented_content.encode("utf-8"),) in result)

        result.addCallback(check)
        return result

    def test_set_umask_appropriately(self):
        """
        We should be setting the umask to 0022 before executing a script, and
        restoring it to the previous value when finishing.
        """
        # Get original umask.
        old_umask = os.umask(0)
        os.umask(old_umask)
        mock_umask = self.mocker.replace("os.umask")
        mock_umask(0022)
        self.mocker.result(old_umask)
        mock_umask(old_umask)
        self.mocker.replay()
        result = self.plugin.run_script("/bin/sh", "umask")
        result.addCallback(self.assertEqual, "%04o\n" % old_umask)
        return result

    def test_restore_umask_in_event_of_error(self):
        """
        We set the umask before executing the script, in the event that there's
        an error setting up the script, we want to restore the umask.
        """
        mock_umask = self.mocker.replace("os.umask")
        mock_umask(0022)
        self.mocker.result(0077)
        mock_mkdtemp = self.mocker.replace("tempfile.mkdtemp",
                                           passthrough=False)
        mock_mkdtemp()
        self.mocker.throw(OSError("Fail!"))
        mock_umask(0077)
        self.mocker.replay()
        result = self.plugin.run_script("/bin/sh", "umask",
                                        attachments={u"file1": "some data"})
        return self.assertFailure(result, OSError)

    def test_run_with_attachments(self):
        result = self.plugin.run_script(
            u"/bin/sh",
            u"ls $LANDSCAPE_ATTACHMENTS && cat $LANDSCAPE_ATTACHMENTS/file1",
            attachments={u"file1": "some data"})

        def check(result):
            self.assertEqual(result, "file1\nsome data")

        result.addCallback(check)
        return result

    def test_run_with_attachment_ids(self):
        """
        The most recent protocol for script message doesn't include the
        attachment body inside the message itself, but instead gives an
        attachment ID, and the plugin fetches the files separately.
        """
        self.manager.config.url = "https://localhost/message-system"
        persist = Persist(
            filename=os.path.join(self.config.data_path, "broker.bpickle"))
        registration_persist = persist.root_at("registration")
        registration_persist.set("secure-id", "secure_id")
        persist.save()
        mock_fetch = self.mocker.replace("landscape.lib.fetch.fetch_async",
                                         passthrough=False)
        headers = {"User-Agent": "landscape-client/%s" % VERSION,
                   "Content-Type": "application/octet-stream",
                   "X-Computer-ID": "secure_id"}
        mock_fetch("https://localhost/attachment/14", headers=headers,
                   cainfo=None)
        self.mocker.result(succeed("some other data"))
        self.mocker.replay()

        result = self.plugin.run_script(
            u"/bin/sh",
            u"ls $LANDSCAPE_ATTACHMENTS && cat $LANDSCAPE_ATTACHMENTS/file1",
            attachments={u"file1": 14})

        def check(result):
            self.assertEqual(result, "file1\nsome other data")

        result.addCallback(check)
        return result

    def test_run_with_attachment_ids_and_ssl(self):
        """
        When fetching attachments, L{ScriptExecution} passes the optional ssl
        certificate file if the configuration specifies it.
        """
        self.manager.config.url = "https://localhost/message-system"
        self.manager.config.ssl_public_key = "/some/key"
        persist = Persist(
            filename=os.path.join(self.config.data_path, "broker.bpickle"))
        registration_persist = persist.root_at("registration")
        registration_persist.set("secure-id", "secure_id")
        persist.save()
        mock_fetch = self.mocker.replace("landscape.lib.fetch.fetch_async",
                                         passthrough=False)
        headers = {"User-Agent": "landscape-client/%s" % VERSION,
                   "Content-Type": "application/octet-stream",
                   "X-Computer-ID": "secure_id"}
        mock_fetch("https://localhost/attachment/14", headers=headers,
                   cainfo="/some/key")
        self.mocker.result(succeed("some other data"))
        self.mocker.replay()

        result = self.plugin.run_script(
            u"/bin/sh",
            u"ls $LANDSCAPE_ATTACHMENTS && cat $LANDSCAPE_ATTACHMENTS/file1",
            attachments={u"file1": 14})

        def check(result):
            self.assertEqual(result, "file1\nsome other data")

        result.addCallback(check)
        return result

    def test_self_remove_script(self):
        """
        If a script removes itself, it doesn't create an error when the script
        execution plugin tries to remove the script file.
        """
        result = self.plugin.run_script("/bin/sh", "echo hi && rm $0")
        result.addCallback(self.assertEqual, "hi\n")
        return result

    def test_self_remove_attachments(self):
        """
        If a script removes its attachments, it doesn't create an error when
        the script execution plugin tries to remove the attachments directory.
        """
        result = self.plugin.run_script(
            u"/bin/sh",
            u"ls $LANDSCAPE_ATTACHMENTS && rm -r $LANDSCAPE_ATTACHMENTS",
            attachments={u"file1": "some data"})

        def check(result):
            self.assertEqual(result, "file1\n")

        result.addCallback(check)
        return result

    def _run_script(self, username, uid, gid, path):
        # ignore the call to chown!
        mock_chown = self.mocker.replace("os.chown", passthrough=False)
        mock_chown(ARGS)

        expected_uid = uid if uid != os.getuid() else None
        expected_gid = gid if gid != os.getgid() else None

        factory = StubProcessFactory()
        self.plugin.process_factory = factory

        self.mocker.replay()

        result = self.plugin.run_script("/bin/sh", "echo hi", user=username)

        self.assertEqual(len(factory.spawns), 1)
        spawn = factory.spawns[0]
        self.assertEqual(spawn[4], path)
        self.assertEqual(spawn[5], expected_uid)
        self.assertEqual(spawn[6], expected_gid)
        result.addCallback(self.assertEqual, "foobar")

        protocol = spawn[0]
        protocol.childDataReceived(1, "foobar")
        for fd in (0, 1, 2):
            protocol.childConnectionLost(fd)
        protocol.processEnded(Failure(ProcessDone(0)))
        return result

    def test_user(self):
        """
        Running a script as a particular user calls
        C{IReactorProcess.spawnProcess} with an appropriate C{uid} argument,
        with the user's primary group as the C{gid} argument and with the user
        home as C{path} argument.
        """
        uid = os.getuid()
        info = pwd.getpwuid(uid)
        username = info.pw_name
        gid = info.pw_gid
        path = info.pw_dir

        return self._run_script(username, uid, gid, path)

    def test_user_no_home(self):
        """
        When the user specified to C{run_script} doesn't have a home, the
        script executes in '/'.
        """
        mock_getpwnam = self.mocker.replace("pwd.getpwnam", passthrough=False)

        class pwnam(object):
            pw_uid = 1234
            pw_gid = 5678
            pw_dir = self.makeFile()

        self.expect(mock_getpwnam("user")).result(pwnam)

        return self._run_script("user", 1234, 5678, "/")

    def test_user_with_attachments(self):
        uid = os.getuid()
        info = pwd.getpwuid(uid)
        username = info.pw_name
        gid = info.pw_gid

        mock_chown = self.mocker.replace("os.chown", passthrough=False)
        mock_chown(ANY, uid, gid)
        self.mocker.count(3)

        factory = StubProcessFactory()
        self.plugin.process_factory = factory

        self.mocker.replay()

        result = self.plugin.run_script("/bin/sh", "echo hi", user=username,
            attachments={u"file 1": "some data"})

        self.assertEqual(len(factory.spawns), 1)
        spawn = factory.spawns[0]
        self.assertIn("LANDSCAPE_ATTACHMENTS", spawn[3])
        attachment_dir = spawn[3]["LANDSCAPE_ATTACHMENTS"]
        self.assertEqual(stat.S_IMODE(os.stat(attachment_dir).st_mode), 0700)
        filename = os.path.join(attachment_dir, "file 1")
        self.assertEqual(stat.S_IMODE(os.stat(filename).st_mode), 0600)

        protocol = spawn[0]
        protocol.childDataReceived(1, "foobar")
        for fd in (0, 1, 2):
            protocol.childConnectionLost(fd)
        protocol.processEnded(Failure(ProcessDone(0)))

        def check(data):
            self.assertEqual(data, "foobar")
            self.assertFalse(os.path.exists(attachment_dir))

        return result.addCallback(check)

    def test_limit_size(self):
        """Data returned from the command is limited."""
        factory = StubProcessFactory()
        self.plugin.process_factory = factory
        self.plugin.size_limit = 100
        result = self.plugin.run_script("/bin/sh", "")
        result.addCallback(self.assertEqual, "x" * 100)

        protocol = factory.spawns[0][0]
        protocol.childDataReceived(1, "x" * 200)
        for fd in (0, 1, 2):
            protocol.childConnectionLost(fd)
        protocol.processEnded(Failure(ProcessDone(0)))

        return result

    def test_limit_time(self):
        """
        The process only lasts for a certain number of seconds.
        """
        result = self.plugin.run_script("/bin/sh", "cat", time_limit=500)
        self.manager.reactor.advance(501)
        self.assertFailure(result, ProcessTimeLimitReachedError)
        return result

    def test_limit_time_accumulates_data(self):
        """
        Data from processes that time out should still be accumulated and
        available from the exception object that is raised.
        """
        factory = StubProcessFactory()
        self.plugin.process_factory = factory
        result = self.plugin.run_script("/bin/sh", "", time_limit=500)
        protocol = factory.spawns[0][0]
        protocol.makeConnection(DummyProcess())
        protocol.childDataReceived(1, "hi\n")
        self.manager.reactor.advance(501)
        protocol.processEnded(Failure(ProcessDone(0)))

        def got_error(f):
            self.assertTrue(f.check(ProcessTimeLimitReachedError))
            self.assertEqual(f.value.data, "hi\n")

        result.addErrback(got_error)
        return result

    def test_time_limit_canceled_after_success(self):
        """
        The timeout call is cancelled after the script terminates.
        """
        factory = StubProcessFactory()
        self.plugin.process_factory = factory
        self.plugin.run_script("/bin/sh", "", time_limit=500)
        protocol = factory.spawns[0][0]
        transport = DummyProcess()
        protocol.makeConnection(transport)
        protocol.childDataReceived(1, "hi\n")
        protocol.processEnded(Failure(ProcessDone(0)))
        self.manager.reactor.advance(501)
        self.assertEqual(transport.signals, [])

    def test_cancel_doesnt_blow_after_success(self):
        """
        When the process ends successfully and is immediately followed by the
        timeout, the output should still be in the failure and nothing bad will
        happen!
        [regression test: killing of the already-dead process would blow up.]
        """
        factory = StubProcessFactory()
        self.plugin.process_factory = factory
        result = self.plugin.run_script("/bin/sh", "", time_limit=500)
        protocol = factory.spawns[0][0]
        protocol.makeConnection(DummyProcess())
        protocol.childDataReceived(1, "hi")
        protocol.processEnded(Failure(ProcessDone(0)))
        self.manager.reactor.advance(501)

        def got_result(output):
            self.assertEqual(output, "hi")

        result.addCallback(got_result)
        return result

    def test_script_is_owned_by_user(self):
        """
        This is a very white-box test. When a script is generated, it must be
        created such that data NEVER gets into it before the file has the
        correct permissions. Therefore os.chmod and os.chown must be called
        before data is written.
        """
        username = pwd.getpwuid(os.getuid())[0]
        uid, gid, home = get_user_info(username)

        mock_chown = self.mocker.replace("os.chown", passthrough=False)
        mock_chmod = self.mocker.replace("os.chmod", passthrough=False)
        mock_mkstemp = self.mocker.replace("tempfile.mkstemp",
                                           passthrough=False)
        mock_fdopen = self.mocker.replace("os.fdopen", passthrough=False)
        process_factory = self.mocker.mock()
        self.plugin.process_factory = process_factory

        self.mocker.order()

        self.expect(mock_mkstemp()).result((99, "tempo!"))

        script_file = mock_fdopen(99, "w")
        mock_chmod("tempo!", 0700)
        mock_chown("tempo!", uid, gid)
        # The contents are written *after* the permissions have been set up!
        script_file.write("#!/bin/sh\ncode")
        script_file.close()
        process_factory.spawnProcess(
            ANY, ANY, uid=None, gid=None, path=ANY,
            env=get_default_environment())
        self.mocker.replay()
        # We don't really care about the deferred that's returned, as long as
        # those things happened in the correct order.
        self.plugin.run_script("/bin/sh", "code",
                               user=pwd.getpwuid(uid)[0])

    def test_script_removed(self):
        """
        The script is removed after it is finished.
        """
        mock_mkstemp = self.mocker.replace("tempfile.mkstemp",
                                           passthrough=False)
        fd, filename = tempfile.mkstemp()
        self.expect(mock_mkstemp()).result((fd, filename))
        self.mocker.replay()
        d = self.plugin.run_script("/bin/sh", "true")
        d.addCallback(lambda ign: self.assertFalse(os.path.exists(filename)))
        return d

    def test_unknown_interpreter(self):
        """
        If the script is run with an unknown interpreter, it raises a
        meaningful error instead of crashing in execvpe.
        """
        d = self.plugin.run_script("/bin/cantpossiblyexist", "stuff")

        def cb(ignore):
            self.fail("Should not be there")

        def eb(failure):
            failure.trap(UnknownInterpreterError)
            self.assertEqual(
                failure.value.interpreter,
                "/bin/cantpossiblyexist")
        return d.addCallback(cb).addErrback(eb)


class ScriptExecutionMessageTests(LandscapeTest):
    helpers = [ManagerHelper]

    def setUp(self):
        super(ScriptExecutionMessageTests, self).setUp()
        self.broker_service.message_store.set_accepted_types(
            ["operation-result"])
        self.manager.config.script_users = "ALL"

    def _verify_script(self, executable, interp, code):
        """
        Given spawnProcess arguments, check to make sure that the temporary
        script has the correct content.
        """
        data = open(executable, "r").read()
        self.assertEqual(data, "#!%s\n%s" % (interp, code))

    def _send_script(self, interpreter, code, operation_id=123,
                     user=pwd.getpwuid(os.getuid())[0],
                     time_limit=None, attachments={},
                     server_supplied_env=None):
        message = {"type": "execute-script",
                   "interpreter": interpreter,
                   "code": code,
                   "operation-id": operation_id,
                   "username": user,
                   "time-limit": time_limit,
                   "attachments": dict(attachments)}
        if server_supplied_env:
            message["env"] = server_supplied_env
        return self.manager.dispatch_message(message)

    def test_success(self):
        """
        When a C{execute-script} message is received from the server, the
        specified script will be run and an operation-result will be sent back
        to the server.
        """
        # Let's use a stub process factory, because otherwise we don't have
        # access to the deferred.
        factory = StubProcessFactory()

        # ignore the call to chown!
        mock_chown = self.mocker.replace("os.chown", passthrough=False)
        mock_chown(ARGS)

        self.manager.add(ScriptExecutionPlugin(process_factory=factory))

        self.mocker.replay()
        result = self._send_script(sys.executable, "print 'hi'")

        self._verify_script(factory.spawns[0][1], sys.executable, "print 'hi'")
        self.assertMessages(
            self.broker_service.message_store.get_pending_messages(), [])

        # Now let's simulate the completion of the process
        factory.spawns[0][0].childDataReceived(1, "hi!\n")
        factory.spawns[0][0].processEnded(Failure(ProcessDone(0)))

        def got_result(r):
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"type": "operation-result",
                  "operation-id": 123,
                  "status": SUCCEEDED,
                  "result-text": u"hi!\n"}])

        result.addCallback(got_result)
        return result

    def test_success_with_server_supplied_env(self):
        """
        When a C{execute-script} message is received from the server, the
        specified script will be run with the supplied environment and an
        operation-result will be sent back to the server.
        """
        # Let's use a stub process factory, because otherwise we don't have
        # access to the deferred.
        factory = StubProcessFactory()

        # ignore the call to chown!
        mock_chown = self.mocker.replace("os.chown", passthrough=False)
        mock_chown(ARGS)

        self.manager.add(ScriptExecutionPlugin(process_factory=factory))

        self.mocker.replay()
        result = self._send_script(sys.executable, "print 'hi'",
                                   server_supplied_env={"Dog": "Woof"})

        self._verify_script(factory.spawns[0][1], sys.executable, "print 'hi'")
        # Verify environment was passed
        self.assertIn("HOME", factory.spawns[0][3])
        self.assertIn("USER", factory.spawns[0][3])
        self.assertIn("PATH", factory.spawns[0][3])
        self.assertIn("Dog", factory.spawns[0][3])
        self.assertEqual("Woof", factory.spawns[0][3]["Dog"])

        self.assertMessages(
            self.broker_service.message_store.get_pending_messages(), [])

        # Now let's simulate the completion of the process
        factory.spawns[0][0].childDataReceived(1, "Woof\n")
        factory.spawns[0][0].processEnded(Failure(ProcessDone(0)))

        def got_result(r):
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"type": "operation-result",
                  "operation-id": 123,
                  "status": SUCCEEDED,
                  "result-text": u"Woof\n"}])

        result.addCallback(got_result)
        return result

    def test_user(self):
        """A user can be specified in the message."""
        username = pwd.getpwuid(os.getuid())[0]
        uid, gid, home = get_user_info(username)

        # ignore the call to chown!
        mock_chown = self.mocker.replace("os.chown", passthrough=False)
        mock_chown(ARGS)

        def spawn_called(protocol, filename, uid, gid, path, env):
            protocol.childDataReceived(1, "hi!\n")
            protocol.processEnded(Failure(ProcessDone(0)))
            self._verify_script(filename, sys.executable, "print 'hi'")

        process_factory = self.mocker.mock()
        process_factory.spawnProcess(
            ANY, ANY, uid=None, gid=None, path=ANY,
            env=get_default_environment())
        self.mocker.call(spawn_called)
        self.mocker.replay()

        self.manager.add(
            ScriptExecutionPlugin(process_factory=process_factory))

        result = self._send_script(sys.executable, "print 'hi'", user=username)
        return result

    def test_unknown_user_with_unicode(self):
        """
        If an error happens because an unknow user is selected, and that this
        user name happens to contain unicode characters, the error message is
        correctly encoded and reported.

        This test mainly ensures that unicode error message works, using
        unknown user as an easy way to test it.
        """
        self.log_helper.ignore_errors(UnknownUserError)
        username = u"non-existent-f\N{LATIN SMALL LETTER E WITH ACUTE}e"
        self.manager.add(
            ScriptExecutionPlugin())

        self._send_script(sys.executable, "print 'hi'", user=username)
        self.assertMessages(
            self.broker_service.message_store.get_pending_messages(),
            [{"type": "operation-result",
              "operation-id": 123,
              "result-text": u"UnknownUserError: Unknown user '%s'" % username,
              "status": FAILED}])

    def test_timeout(self):
        """
        If a L{ProcessTimeLimitReachedError} is fired back, the
        operation-result should have a failed status.
        """
        factory = StubProcessFactory()
        self.manager.add(ScriptExecutionPlugin(process_factory=factory))

        # ignore the call to chown!
        mock_chown = self.mocker.replace("os.chown", passthrough=False)
        mock_chown(ARGS)

        self.mocker.replay()
        result = self._send_script(sys.executable, "bar", time_limit=30)
        self._verify_script(factory.spawns[0][1], sys.executable, "bar")

        protocol = factory.spawns[0][0]
        protocol.makeConnection(DummyProcess())
        protocol.childDataReceived(2, "ONOEZ")
        self.manager.reactor.advance(31)
        protocol.processEnded(Failure(ProcessDone(0)))

        def got_result(r):
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"type": "operation-result",
                  "operation-id": 123,
                  "status": FAILED,
                  "result-text": u"ONOEZ",
                  "result-code": 102}])

        result.addCallback(got_result)
        return result

    def test_configured_users(self):
        """
        Messages which try to run a script as a user that is not allowed should
        be rejected.
        """
        self.manager.add(ScriptExecutionPlugin())
        self.manager.config.script_users = "landscape, nobody"
        result = self._send_script(sys.executable, "bar", user="whatever")

        def got_result(r):
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"type": "operation-result",
                  "operation-id": 123,
                  "status": FAILED,
                  "result-text": u"Scripts cannot be run as user whatever."}])

        result.addCallback(got_result)
        return result

    def test_urgent_response(self):
        """Responses to script execution messages are urgent."""
        # ignore the call to chown!
        mock_chown = self.mocker.replace("os.chown", passthrough=False)
        mock_chown(ARGS)

        def spawn_called(protocol, filename, uid, gid, path, env):
            protocol.childDataReceived(1, "hi!\n")
            protocol.processEnded(Failure(ProcessDone(0)))
            self._verify_script(filename, sys.executable, "print 'hi'")

        process_factory = self.mocker.mock()
        process_factory.spawnProcess(
            ANY, ANY, uid=None, gid=None, path=ANY,
            env=get_default_environment())
        self.mocker.call(spawn_called)

        self.mocker.replay()

        self.manager.add(
            ScriptExecutionPlugin(process_factory=process_factory))

        def got_result(r):
            self.assertTrue(self.broker_service.exchanger.is_urgent())
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"type": "operation-result",
                  "operation-id": 123,
                  "result-text": u"hi!\n",
                  "status": SUCCEEDED}])

        result = self._send_script(sys.executable, "print 'hi'")
        result.addCallback(got_result)
        return result

    def test_binary_output(self):
        """
        If a script outputs non-printable characters not handled by utf-8, they
        are replaced during the encoding phase but the script succeeds.
        """
        mock_chown = self.mocker.replace("os.chown", passthrough=False)
        mock_chown(ARGS)

        def spawn_called(protocol, filename, uid, gid, path, env):
            protocol.childDataReceived(1,
            "\x7fELF\x01\x01\x01\x00\x00\x00\x95\x01")
            protocol.processEnded(Failure(ProcessDone(0)))
            self._verify_script(filename, sys.executable, "print 'hi'")

        process_factory = self.mocker.mock()
        process_factory.spawnProcess(
            ANY, ANY, uid=None, gid=None, path=ANY,
            env=get_default_environment())
        self.mocker.call(spawn_called)

        self.mocker.replay()

        self.manager.add(
            ScriptExecutionPlugin(process_factory=process_factory))

        def got_result(r):
            self.assertTrue(self.broker_service.exchanger.is_urgent())
            [message] = (
                self.broker_service.message_store.get_pending_messages())
            self.assertEqual(
                message["result-text"],
                 u"\x7fELF\x01\x01\x01\x00\x00\x00\ufffd\x01")

        result = self._send_script(sys.executable, "print 'hi'")
        result.addCallback(got_result)
        return result

    def test_parse_error_causes_operation_failure(self):
        """
        If there is an error parsing the message, an operation-result will be
        sent (assuming operation-id *is* successfully parsed).
        """
        self.log_helper.ignore_errors(KeyError)
        self.manager.add(ScriptExecutionPlugin())

        self.manager.dispatch_message(
            {"type": "execute-script", "operation-id": 444})

        expected_message = [{"type": "operation-result",
                             "operation-id": 444,
                             "result-text": u"KeyError: username",
                             "status": FAILED}]

        self.assertMessages(
            self.broker_service.message_store.get_pending_messages(),
            expected_message)

        self.assertTrue("KeyError: 'username'" in self.logfile.getvalue())

    def test_non_zero_exit_fails_operation(self):
        """
        If a script exits with a nen-zero exit code, the operation associated
        with it should fail, but the data collected should still be sent.
        """
        # Mock a bunch of crap so that we can run a real process
        self.mocker.replace("os.chown", passthrough=False)(ARGS)
        self.mocker.replace("os.setuid", passthrough=False)(ARGS)
        self.mocker.count(0, None)
        self.mocker.replace("os.setgid", passthrough=False)(ARGS)
        self.mocker.count(0, None)
        self.mocker.count(0, None)
        self.mocker.replay()

        self.manager.add(ScriptExecutionPlugin())
        result = self._send_script("/bin/sh", "echo hi; exit 1")

        def got_result(ignored):
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"type": "operation-result",
                  "operation-id": 123,
                  "result-text": "hi\n",
                  "result-code": PROCESS_FAILED_RESULT,
                  "status": FAILED}])

        return result.addCallback(got_result)

    def test_unknown_error(self):
        """
        When a completely unknown error comes back from the process protocol,
        the operation fails and the formatted failure is included in the
        response message.
        """
        factory = StubProcessFactory()

        # ignore the call to chown!
        mock_chown = self.mocker.replace("os.chown", passthrough=False)
        mock_chown(ARGS)

        self.manager.add(ScriptExecutionPlugin(process_factory=factory))

        self.mocker.replay()
        result = self._send_script(sys.executable, "print 'hi'")

        self._verify_script(factory.spawns[0][1], sys.executable, "print 'hi'")
        self.assertMessages(
            self.broker_service.message_store.get_pending_messages(), [])

        failure = Failure(RuntimeError("Oh noes!"))
        factory.spawns[0][0].result_deferred.errback(failure)

        def got_result(r):
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"type": "operation-result",
                  "operation-id": 123,
                  "status": FAILED,
                  "result-text": str(failure)}])

        result.addCallback(got_result)
        return result

    def test_fetch_attachment_failure(self):
        """
        If the plugin fails to retrieve the attachments with a
        L{HTTPCodeError}, a specific error code is shown.
        """
        self.manager.config.url = "https://localhost/message-system"
        persist = Persist(
            filename=os.path.join(self.config.data_path, "broker.bpickle"))
        registration_persist = persist.root_at("registration")
        registration_persist.set("secure-id", "secure_id")
        persist.save()
        mock_fetch = self.mocker.replace("landscape.lib.fetch.fetch_async",
                                         passthrough=False)
        headers = {"User-Agent": "landscape-client/%s" % VERSION,
                   "Content-Type": "application/octet-stream",
                   "X-Computer-ID": "secure_id"}
        mock_fetch("https://localhost/attachment/14", headers=headers,
                   cainfo=None)
        self.mocker.result(fail(HTTPCodeError(404, "Not found")))
        self.mocker.replay()

        self.manager.add(ScriptExecutionPlugin())
        result = self._send_script(
            "/bin/sh", "echo hi", attachments={u"file1": 14})

        def got_result(ignored):
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"type": "operation-result",
                  "operation-id": 123,
                  "result-text": "Server returned HTTP code 404",
                  "result-code": FETCH_ATTACHMENTS_FAILED_RESULT,
                  "status": FAILED}])

        return result.addCallback(got_result)



"""Provide access to the persistent data used by L{PackageTaskHandler}s."""
import time

try:
    import sqlite3
except ImportError:
    from pysqlite2 import dbapi2 as sqlite3

from landscape.lib import bpickle
from landscape.lib.store import with_cursor


class UnknownHashIDRequest(Exception):
    """Raised for unknown hash id requests."""


class InvalidHashIdDb(Exception):
    """Raised when trying to add an invalid hash=>id lookaside database."""


class HashIdStore(object):
    """C{HashIdStore} stores package hash=>id mappings in a file.

    The file is a SQLite database that contains a single table called "hash".
    The table schema is defined in L{ensure_hash_id_schema}.

    @param filename: The file where the mappings are persisted to.
    """
    _db = None

    def __init__(self, filename):
        self._filename = filename

    def _ensure_schema(self):
        ensure_hash_id_schema(self._db)

    @with_cursor
    def set_hash_ids(self, cursor, hash_ids):
        """Set the ids of a set of hashes.

        @param hash_ids: a C{dict} of hash=>id mappings.
        """
        for hash, id in hash_ids.iteritems():
            cursor.execute("REPLACE INTO hash VALUES (?, ?)",
                           (id, buffer(hash)))

    @with_cursor
    def get_hash_id(self, cursor, hash):
        """Return the id associated to C{hash}, or C{None} if not available."""
        cursor.execute("SELECT id FROM hash WHERE hash=?", (buffer(hash),))
        value = cursor.fetchone()
        if value:
            return value[0]
        return None

    @with_cursor
    def get_hash_ids(self, cursor):
        """Return a C{dict} holding all the available hash=>id mappings."""
        cursor.execute("SELECT hash, id FROM hash")
        return dict([(str(row[0]), row[1]) for row in cursor.fetchall()])

    @with_cursor
    def get_id_hash(self, cursor, id):
        """Return the hash associated to C{id}, or C{None} if not available."""
        assert isinstance(id, (int, long))
        cursor.execute("SELECT hash FROM hash WHERE id=?", (id,))
        value = cursor.fetchone()
        if value:
            return str(value[0])
        return None

    @with_cursor
    def clear_hash_ids(self, cursor):
        """Delete all hash=>id mappings."""
        cursor.execute("DELETE FROM hash")

    @with_cursor
    def check_sanity(self, cursor):
        """Check database integrity.

        @raise: L{InvalidHashIdDb} if the filenme passed to the constructor is
            not a SQLite database or does not have a table called "hash" with
            a compatible schema.
        """
        try:
            cursor.execute("SELECT id FROM hash WHERE hash=?", ("",))
        except sqlite3.DatabaseError:
            raise InvalidHashIdDb(self._filename)


class PackageStore(HashIdStore):
    """Persist data about system packages and L{PackageTaskHandler}'s tasks.

    This class extends L{HashIdStore} by adding tables to the SQLite database
    backend for storing information about the status of the system packages and
    about the tasks to be performed by L{PackageTaskHandler}s.

    The additional tables and schemas are defined in L{ensure_package_schema}.

    @param filename: The file where data is persisted to.
    """

    def __init__(self, filename):
        super(PackageStore, self).__init__(filename)
        self._hash_id_stores = []

    def _ensure_schema(self):
        super(PackageStore, self)._ensure_schema()
        ensure_package_schema(self._db)

    def add_hash_id_db(self, filename):
        """
        Attach an additional "lookaside" hash=>id database.

        This method can be called more than once to attach several
        hash=>id databases, which will be queried *before* the main
        database, in the same the order they were added.

        If C{filename} is not a SQLite database or does not have a
        table called "hash" with a compatible schema, L{InvalidHashIdDb}
        is raised.

        @param filename: a secondary SQLite databases to look for pre-canned
                         hash=>id mappings.
        """
        hash_id_store = HashIdStore(filename)

        try:
            hash_id_store.check_sanity()
        except InvalidHashIdDb, e:
            # propagate the error
            raise e

        self._hash_id_stores.append(hash_id_store)

    def has_hash_id_db(self):
        """Return C{True} if one or more lookaside databases are attached."""
        return len(self._hash_id_stores) > 0

    def get_hash_id(self, hash):
        """Return the id associated to C{hash}, or C{None} if not available.

        This method composes the L{HashIdStore.get_hash_id} methods of all
        the attached lookaside databases, falling back to the main one, as
        described in L{add_hash_id_db}.
        """
        assert isinstance(hash, basestring)

        # Check if we can find the hash=>id mapping in the lookaside stores
        for store in self._hash_id_stores:
            id = store.get_hash_id(hash)
            if id:
                return id

        # Fall back to the locally-populated db
        return HashIdStore.get_hash_id(self, hash)

    def get_id_hash(self, id):
        """Return the hash associated to C{id}, or C{None} if not available.

        This method composes the L{HashIdStore.get_id_hash} methods of all
        the attached lookaside databases, falling back to the main one in
        case the hash associated to C{id} is not found in any of them.
        """
        for store in self._hash_id_stores:
            hash = store.get_id_hash(id)
            if hash is not None:
                return hash
        return HashIdStore.get_id_hash(self, id)

    @with_cursor
    def add_available(self, cursor, ids):
        for id in ids:
            cursor.execute("REPLACE INTO available VALUES (?)", (id,))

    @with_cursor
    def remove_available(self, cursor, ids):
        id_list = ",".join(str(int(id)) for id in ids)
        cursor.execute("DELETE FROM available WHERE id IN (%s)" % id_list)

    @with_cursor
    def clear_available(self, cursor):
        cursor.execute("DELETE FROM available")

    @with_cursor
    def get_available(self, cursor):
        cursor.execute("SELECT id FROM available")
        return [row[0] for row in cursor.fetchall()]

    @with_cursor
    def add_available_upgrades(self, cursor, ids):
        for id in ids:
            cursor.execute("REPLACE INTO available_upgrade VALUES (?)", (id,))

    @with_cursor
    def remove_available_upgrades(self, cursor, ids):
        id_list = ",".join(str(int(id)) for id in ids)
        cursor.execute("DELETE FROM available_upgrade WHERE id IN (%s)"
                       % id_list)

    @with_cursor
    def clear_available_upgrades(self, cursor):
        cursor.execute("DELETE FROM available_upgrade")

    @with_cursor
    def get_available_upgrades(self, cursor):
        cursor.execute("SELECT id FROM available_upgrade")
        return [row[0] for row in cursor.fetchall()]

    @with_cursor
    def add_installed(self, cursor, ids):
        for id in ids:
            cursor.execute("REPLACE INTO installed VALUES (?)", (id,))

    @with_cursor
    def remove_installed(self, cursor, ids):
        id_list = ",".join(str(int(id)) for id in ids)
        cursor.execute("DELETE FROM installed WHERE id IN (%s)" % id_list)

    @with_cursor
    def clear_installed(self, cursor):
        cursor.execute("DELETE FROM installed")

    @with_cursor
    def get_installed(self, cursor):
        cursor.execute("SELECT id FROM installed")
        return [row[0] for row in cursor.fetchall()]

    @with_cursor
    def get_locked(self, cursor):
        """Get the package ids of all locked packages."""
        cursor.execute("SELECT id FROM locked")
        return [row[0] for row in cursor.fetchall()]

    @with_cursor
    def add_locked(self, cursor, ids):
        """Add the given package ids to the list of locked packages."""
        for id in ids:
            cursor.execute("REPLACE INTO locked VALUES (?)", (id,))

    @with_cursor
    def remove_locked(self, cursor, ids):
        id_list = ",".join(str(int(id)) for id in ids)
        cursor.execute("DELETE FROM locked WHERE id IN (%s)" % id_list)

    @with_cursor
    def clear_locked(self, cursor):
        """Remove all the package ids in the locked table."""
        cursor.execute("DELETE FROM locked")

    @with_cursor
    def add_hash_id_request(self, cursor, hashes):
        hashes = list(hashes)
        cursor.execute("INSERT INTO hash_id_request (hashes, timestamp)"
                       " VALUES (?,?)",
                       (buffer(bpickle.dumps(hashes)), time.time()))
        return HashIDRequest(self._db, cursor.lastrowid)

    @with_cursor
    def get_hash_id_request(self, cursor, request_id):
        cursor.execute("SELECT 1 FROM hash_id_request WHERE id=?",
                       (request_id,))
        if not cursor.fetchone():
            raise UnknownHashIDRequest(request_id)
        return HashIDRequest(self._db, request_id)

    @with_cursor
    def iter_hash_id_requests(self, cursor):
        cursor.execute("SELECT id FROM hash_id_request")
        return [HashIDRequest(self._db, row[0]) for row in cursor.fetchall()]

    @with_cursor
    def clear_hash_id_requests(self, cursor):
        cursor.execute("DELETE FROM hash_id_request")

    @with_cursor
    def add_task(self, cursor, queue, data):
        data = bpickle.dumps(data)
        cursor.execute("INSERT INTO task (queue, timestamp, data) "
                       "VALUES (?,?,?)", (queue, time.time(), buffer(data)))
        return PackageTask(self._db, cursor.lastrowid)

    @with_cursor
    def get_next_task(self, cursor, queue):
        cursor.execute("SELECT id FROM task WHERE queue=? ORDER BY timestamp",
                       (queue,))
        row = cursor.fetchone()
        if row:
            return PackageTask(self._db, row[0])
        return None

    @with_cursor
    def clear_tasks(self, cursor, except_tasks=()):
        cursor.execute("DELETE FROM task WHERE id NOT IN (%s)" %
                       ",".join([str(task.id) for task in except_tasks]))


class FakePackageStore(PackageStore):
    """
    A L{PackageStore} with an additional message table to store sent messages.
    """

    def _ensure_schema(self):
        super(FakePackageStore, self)._ensure_schema()
        ensure_fake_package_schema(self._db)

    @with_cursor
    def save_message(self, cursor, message):
        cursor.execute("INSERT INTO message (data) VALUES (?)",
                       (buffer(bpickle.dumps(message)),))

    @with_cursor
    def get_message_ids(self, cursor):
        return [row[0] for row in
                cursor.execute("SELECT id FROM message").fetchall()]

    @with_cursor
    def save_message_ids(self, cursor, message_ids):
        cursor.executemany(
            "INSERT INTO message (id) VALUES (?)",
            [(message_id,) for message_id in message_ids])

    @with_cursor
    def get_messages_by_ids(self, cursor, message_ids):
        params = ", ".join(["?"] * len(message_ids))
        result = cursor.execute(
                "SELECT id, data FROM message WHERE id IN (%s) "
                "ORDER BY id" % params, tuple(message_ids)).fetchall()
        return [(row[0], row[1]) for row in result]


class HashIDRequest(object):

    def __init__(self, db, id):
        self._db = db
        self.id = id

    @property
    @with_cursor
    def hashes(self, cursor):
        cursor.execute("SELECT hashes FROM hash_id_request WHERE id=?",
                       (self.id,))
        return bpickle.loads(str(cursor.fetchone()[0]))

    @with_cursor
    def _get_timestamp(self, cursor):
        cursor.execute("SELECT timestamp FROM hash_id_request WHERE id=?",
                       (self.id,))
        return cursor.fetchone()[0]

    @with_cursor
    def _set_timestamp(self, cursor, value):
        cursor.execute("UPDATE hash_id_request SET timestamp=? WHERE id=?",
                       (value, self.id))

    timestamp = property(_get_timestamp, _set_timestamp)

    @with_cursor
    def _get_message_id(self, cursor):
        cursor.execute("SELECT message_id FROM hash_id_request WHERE id=?",
                       (self.id,))
        return cursor.fetchone()[0]

    @with_cursor
    def _set_message_id(self, cursor, value):
        cursor.execute("UPDATE hash_id_request SET message_id=? WHERE id=?",
                       (value, self.id))

    message_id = property(_get_message_id, _set_message_id)

    @with_cursor
    def remove(self, cursor):
        cursor.execute("DELETE FROM hash_id_request WHERE id=?", (self.id,))


class PackageTask(object):

    def __init__(self, db, id):
        self._db = db
        self.id = id

        cursor = db.cursor()
        try:
            cursor.execute("SELECT queue, timestamp, data FROM task "
                           "WHERE id=?", (id,))
            row = cursor.fetchone()
        finally:
            cursor.close()

        self.queue = row[0]
        self.timestamp = row[1]
        self.data = bpickle.loads(str(row[2]))

    @with_cursor
    def remove(self, cursor):
        cursor.execute("DELETE FROM task WHERE id=?", (self.id,))


def ensure_hash_id_schema(db):
    """Create all tables needed by a L{HashIdStore}.

    @param db: A connection to a SQLite database.
    """
    cursor = db.cursor()
    try:
        cursor.execute("CREATE TABLE hash"
                       " (id INTEGER PRIMARY KEY, hash BLOB UNIQUE)")
    except (sqlite3.OperationalError, sqlite3.DatabaseError):
        cursor.close()
        db.rollback()
    else:
        cursor.close()
        db.commit()


def ensure_package_schema(db):
    """Create all tables needed by a L{PackageStore}.

    @param db: A connection to a SQLite database.
    """
    # FIXME This needs a "patch" table with a "version" column which will
    #       help with upgrades.  It should also be used to decide when to
    #       create the schema from the ground up, rather than that using
    #       try block.
    cursor = db.cursor()
    try:
        cursor.execute("CREATE TABLE locked"
                       " (id INTEGER PRIMARY KEY)")
        cursor.execute("CREATE TABLE available"
                       " (id INTEGER PRIMARY KEY)")
        cursor.execute("CREATE TABLE available_upgrade"
                       " (id INTEGER PRIMARY KEY)")
        cursor.execute("CREATE TABLE installed"
                       " (id INTEGER PRIMARY KEY)")
        cursor.execute("CREATE TABLE hash_id_request"
                       " (id INTEGER PRIMARY KEY, timestamp TIMESTAMP,"
                       " message_id INTEGER, hashes BLOB)")
        cursor.execute("CREATE TABLE task"
                       " (id INTEGER PRIMARY KEY, queue TEXT,"
                       " timestamp TIMESTAMP, data BLOB)")
    except sqlite3.OperationalError:
        cursor.close()
        db.rollback()
    else:
        cursor.close()
        db.commit()


def ensure_fake_package_schema(db):
    cursor = db.cursor()
    try:
        cursor.execute("CREATE TABLE message"
                       " (id INTEGER PRIMARY KEY, data BLOB)")
    except (sqlite3.OperationalError, sqlite3.DatabaseError):
        cursor.close()
        db.rollback()
    else:
        cursor.close()
        db.commit()



from landscape.user.management import UserManagementError
from landscape.user.provider import UserProviderBase


class FakeUserManagement(object):

    def __init__(self, provider=None):
        self.shadow_file = getattr(provider, "shadow_file", None)
        self.provider = provider
        self._users = {}

        for data in self.provider.get_users():
            self._users[data["username"]] = data
        self._groups = {}
        for data in self.provider.get_groups():
            self._groups[data["name"]] = data

    def _make_fake_shadow_file(self, locked_users, unlocked_users):
        entry = "%s:%s:13348:0:99999:7:::\n"
        shadow_file = open(self.shadow_file, "w")
        for user in locked_users:
            shadow_file.write(entry % (user, "!"))
        for user in unlocked_users:
            shadow_file.write(entry % (user, "qweqweqeqweqw"))
        shadow_file.close()

    def add_user(self, username, name, password, require_password_reset,
                 primary_group_name, location, work_phone, home_phone):
        try:
            uid = 1000
            if self._users:
                uid = max([x["uid"] for x in self._users.itervalues()]) + 1
            if self._groups:
                primary_gid = self.get_gid(primary_group_name)
            else:
                primary_gid = uid
            self._users[uid] = {"username": username, "name": name,
                                "uid": uid, "enabled": True,
                                "location": location, "work-phone": work_phone,
                                "home-phone": home_phone,
                                "primary-gid": primary_gid}
            gecos_string = "%s,%s,%s,%s" % (name, location or "",
                                            work_phone or "", home_phone or "")
            userdata = (username, "x", uid, primary_gid, gecos_string,
                        "/bin/sh" , "/home/user")
            self.provider.users.append(userdata)
        except KeyError:
            raise UserManagementError("add_user failed")
        return "add_user succeeded"

    def lock_user(self, username):
        data = self._users.get(username, None)
        if data:
            data["enabled"] = False
            # This will generate a shadow file with only the locked user in it.
            self._make_fake_shadow_file([username], [])
            return "lock_user succeeded"
        raise UserManagementError("lock_user failed")

    def unlock_user(self, username):
        data = self._users.get(username, None)
        if data:
            data["enabled"] = True
            # This will generate a shadow file with only the unlocked user in it.
            self._make_fake_shadow_file([], [username])
            return "unlock_user succeeded"
        raise UserManagementError("unlock_user failed")

    def remove_user(self, username, delete_home=False):
        try:
            del self._users[username]
        except KeyError:
            raise UserManagementError("remove_user failed")
        remaining_users = []
        for user in self.provider.users:
            if user[0] != username:
                remaining_users.append(user)
        self.provider.users = remaining_users
        return "remove_user succeeded"

    def set_user_details(self, username, password=None, name=None,
                         location=None, work_number=None, home_number=None,
                         primary_group_name=None):
        data = self._users.setdefault(username, {})
        for key, value in [("name", name), ("location", location),
                            ("work-phone", work_number),
                            ("home-phone", home_number)]:
            if value:
                data[key] = value
        if primary_group_name:
            data["primary-gid"] = self.get_gid(primary_group_name)
        else:
            data["primary-gid"] = None
        userdata = (username, "x", data["uid"], data["primary-gid"],
                    "%s,%s,%s,%s," % (name, location, work_number,
                                      home_number),
                    "/bin/sh" , "/home/user")
        self.provider.users = [userdata]
        return "set_user_details succeeded"

    def get_gid(self, name):
        try:
            return self._groups[name]["gid"]
        except KeyError:
            raise UserManagementError("Group %s wasn't found." % name)

    def add_group(self, name):
        gid = 1000
        if self._groups:
            gid = max([x["gid"] for x in self._groups.itervalues()]) + 1
        self._groups[name] = {"name": name, "gid": gid, "members": []}
        self.update_provider_from_groups()
        return "add_group succeeded"

    def set_group_details(self, group, new_name):
        data = self._groups[group]
        data["name"] = new_name
        self._groups[new_name] = data
        del self._groups[group]
        self.update_provider_from_groups()
        return "set_group_details succeeded"

    def add_group_member(self, username, group):
        data = self._groups[group]
        if data:
            data["members"].append(username)
            self.update_provider_from_groups()
            return "add_group_member succeeded"
        raise UserManagementError("add_group_member failed")

    def remove_group_member(self, username, group):
        if group in self._groups:
            data = self._groups[group]
            data["members"].remove(username)
            self.update_provider_from_groups()
            return "remove_group_member succeeded"
        raise UserManagementError("remove_group_member failed")

    def remove_group(self, group):
        del self._groups[group]
        self.update_provider_from_groups()
        return "remove_group succeeded"

    def update_provider_from_groups(self):
        provider_list = []
        for k, v in self._groups.iteritems():
            provider_list.append((k, "x", v["gid"], v["members"]))
        self.provider.groups = provider_list


class FakeUserProvider(UserProviderBase):

    def __init__(self, users=None, groups=None, popen=None, shadow_file=None, 
                 locked_users=None):
        self.users = users
        self.groups = groups
        if popen:
            self.popen = popen
        self.shadow_file = shadow_file
        super(FakeUserProvider, self).__init__(locked_users=locked_users)

    def get_user_data(self, system=False):
        if self.users is None:
            self.users = []
        return self.users

    def get_group_data(self):
        if self.groups is None:
            self.groups = []
        return self.groups

class FakeUserInfo(object):
    """Implements enough functionality to work for Changes tests."""

    persist_name = "users"
    run_interval = 60

    def __init__(self, provider):
        self._provider = provider

    def register(self, manager):
        self._manager = manager
        self._persist = self._manager.persist.root_at("users")
